/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: TypeInfo.java 511423 2007-02-25 03:21:05Z vgritsenko $
 */

package org.w3c.dom;

/**
 * The TypeInfo interface represents a type referenced from <code>Element</code>
 * or <code>Attr</code> nodes, specified in the schemas associated with the
 * document. The type is a pair of a namespace URI and name properties, and
 * depends on the document's schema.
 * <p/>
 *  If the document's schema is an XML DTD [XML 1.0], the values are computed as
 * follows:
 * <ul>
 * <li>If this type is referenced from an <code>Attr</code> node, <code>typeNamespace</code> is
 * "<code>http://www.w3.org/TR/REC-xml</code>" and <code>typeName</code> represents the
 * <b>[attribute type]</b> property in the [XML Information Set]. If there is no declaration
 * for the attribute, <code>typeNamespace</code> and <code>typeName</code> are <code>null</code>.
 * <li>If this type is referenced from an <code>Element</code> node, <code>typeNamespace</code>
 * and <code>typeName</code> are <code>null</code>.
 * </ul>
 * <p/>
 * If the document's schema is an XML Schema [XML Schema Part 1], the values are computed as
 * follows using the post-schema-validation infoset contributions (also called PSVI
 * contributions):
 * <ul>
 * <li> If the <b>[validity]</b> property exists AND is <i>"invalid"</i> or <i>"notKnown"</i>:
 * the {target namespace} and {name} properties of the declared type if
 * available, otherwise <code>null</code>.
 * <p/>
 * <b>Note</b>: At the time of writing, the XML Schema specification does not
 * require exposing the declared type. Thus, DOM implementations might choose
 * not to provide type information if validity is not valid.
 * <li>If the <b>[validity]</b> property exists and is <i>"valid"</i>:
 * <ol>
 * <li>If <b>[member type definition]</b> exists:
 * <ol>
 * <li>If {name} is not absent, then expose {name} and {target namespace}
 * properties of the <b>[member type definition]</b> property;
 * <li>Otherwise, expose the namespace and local name of the corresponding
 * anonymous type name.
 * </ol>
 * <li>If the <b>[type definition]</b> property exists:
 * <ol>
 * <li>If {name} is not absent, then expose {name} and {target namespace}
 * properties of the <b>[type definition]</b> property;
 * <li>Otherwise, expose the namespace and local name of the corresponding
 * anonymous type name.
 * </ol>
 * <li>If the <b>[member type definition anonymous]</b> exists:
 * <ol>
 * <li>If it is false, then expose <b>[member type definition name]</b> and <b>[member
 * type definition namespace]</b> properties;
 * <li>Otherwise, expose the namespace and local name of the corresponding
 * anonymous type name.
 * </ol>
 * <li>If the <b>[type definition anonymous]</b> exists:
 * <ol>
 * <li>If it is false, then expose <b>[type definition name]</b> and <b>[type definition
 * namespace]</b> properties;
 * <li>Otherwise, expose the namespace and local name of the corresponding
 * anonymous type name.
 * </ol>
 * </ol>
 * </ul>
 *
 * @since DOM Level 3
 * @version $Revision: 511423 $, $Date: 2007-02-24 22:21:05 -0500 (Sat, 24 Feb 2007) $
 */
public interface TypeInfo {

    /**
     * If the document's schema is an XML Schema [XML Schema Part 1], this
     * constant represents the derivation by restriction if complex types
     * are involved, or a restriction if simple types are involved.
     * <p/>
     * The reference type definition is derived by restriction from the
     * other type definition if the other type definition is the same as the
     * reference type definition, or if the other type definition can be
     * reached recursively following the {base type definition} property from
     * the reference type definition, and all the derivation methods involved
     * are restriction.
     */
    public static final int DERIVATION_RESTRICTION    = 0x00000001;

    /**
     * If the document's schema is an XML Schema [XML Schema Part 1], this
     * constant represents the derivation by extension.
     * <p/>
     * The reference type definition is derived by extension from the other
     * type definition if the other type definition can be reached recursively
     * following the {base type definition} property from the reference type
     * definition, and at least one of the derivation methods involved is an
     * extension.
     */
    public static final int DERIVATION_EXTENSION      = 0x00000002;

    /**
     * If the document's schema is an XML Schema [XML Schema Part 1], this
     * constant represents the union if simple types are involved.
     * <p/>
     * The reference type definition is derived by union from the other type
     * definition if there exists two type definitions T1 and T2 such as the
     * reference type definition is derived from T1 by DERIVATION_RESTRICTION
     * or DERIVATION_EXTENSION, T2 is derived from the other type definition
     * by DERIVATION_RESTRICTION, T1 has {variety} union, and one of the
     * {member type definitions} is T2. Note that T1 could be the same as the
     * reference type definition, and T2 could be the same as the other type
     * definition.
     */
    public static final int DERIVATION_UNION          = 0x00000004;

    /**
     * If the document's schema is an XML Schema [XML Schema Part 1], this
     * constant represents the list.
     * <p/>
     * The reference type definition is derived by list from the other type
     * definition if there exists two type definitions T1 and T2 such as the
     * reference type definition is derived from T1 by DERIVATION_RESTRICTION
     * or DERIVATION_EXTENSION, T2 is derived from the other type definition
     * by DERIVATION_RESTRICTION, T1 has {variety} list, and T2 is the {item
     * type definition}. Note that T1 could be the same as the reference type
     * definition, and T2 could be the same as the other type definition.
     */
    public static final int DERIVATION_LIST           = 0x00000008;

    /**
     * The name of a type declared for the associated element or attribute, or
     * <code>null</code> if unknown.
     * @return The name of a type declared for the associated element or
     * attribute, or null if unknown.
     */
    public String getTypeName();

    /**
     * The namespace of the type declared for the associated element or
     * attribute or <code>null</code> if the element does not have declaration
     * or if no namespace information is available.
     * @return The namespace of the type declared for the associated element
     * or attribute or null if the element does not have declaration or if
     * no namespace information is available.
     */
    public String getTypeNamespace();

    /**
     * This method returns if there is a derivation between the reference type
     * definition, i.e. the TypeInfo on which the method is being called, and
     * the other type definition, i.e. the one passed as parameters.
     * @param typeNamespaceArg the namespace of the other type definition.
     * @param typeNameArg the name of the other type definition.
     * @param derivationMethod the type of derivation and conditions applied
     * between two types, as described in the list of constants provided in this
     * interface.
     * @return If the document's schema is a DTD or no schema is associated
     * with the document, this method will always return false.
     * <p/>
     * If the document's schema is an XML Schema, the method will true if the
     * reference type definition is derived from the other type definition
     * according to the derivation parameter. If the value of the parameter is
     * 0 (no bit is set to 1 for the derivationMethod parameter), the method
     * will return true if the other type definition can be reached by
     * recursing any combination of {base type definition}, {item type
     * definition}, or {member type definitions} from the reference type
     * definition.
     */
    public boolean isDerivedFrom(String typeNamespaceArg, String typeNameArg, int derivationMethod);
}
