/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: CollectionManager.java 541508 2007-05-25 01:54:12Z vgritsenko $
 */

package org.apache.xindice.client.xmldb.services;

import org.w3c.dom.Document;
import org.xmldb.api.base.Collection;
import org.xmldb.api.base.XMLDBException;

/**
 * CollectionManager provides management facilities for a Collection instance.
 * Administrative access is required for retrieval of the CollectionManager
 * interface. User level access is provided through the underlying Collection
 * itself.
 *
 * @version $Revision: 541508 $, $Date: 2007-05-24 21:54:12 -0400 (Thu, 24 May 2007) $
 */
public interface CollectionManager {

    /**
     * Returns the name of the collection that this manager is associated with.
     *
     * @return the name of the collection
     * @exception XMLDBException
     */
    public String getCollectionName() throws XMLDBException;

    /**
     * Returns the fully qualified name of the collection that this manager is
     * associated with. This name includes all parent collections.
     *
     * @return the fully qualified name for this collection.
     * @exception XMLDBException
     */
    public String getCanonicalName() throws XMLDBException;

    /**
     * Creates a new collection in the database identified by name and using
     * the provided configuration.
     *
     * @param path the path of the new collection
     * @param configuration the XML collection configuration to use for
     *        creating this collection.
     * @return The newly created collection
     * @exception XMLDBException
     */
    public Collection createCollection(String path, Document configuration)
            throws XMLDBException;

    /**
     * Drops a child collection from this collection.
     *
     * @param name The name of the child collection to drop.
     * @exception XMLDBException
     */
    public void dropCollection(String name) throws XMLDBException;

    /**
     * Returns a list of all indexers for this collection.
     *
     * @return the list of indexers
     * @exception XMLDBException
     */
    public String[] listIndexers() throws XMLDBException;

    /**
     * Creates a new Indexer for this collection.
     *
     * @param configuration The configuration to use for this indexer.
     * @exception XMLDBException
     */
    public void createIndexer(Document configuration)
            throws XMLDBException;

    /**
     * Drops the indexer from the collection
     *
     * @param name The name of the indexer to drop.
     * @exception XMLDBException
     */
    public void dropIndexer(String name)
            throws XMLDBException;

    /**
     * Returns a list of all collection level XMLObjects for this collection.
     *
     * @return the list of XMLObjects.
     * @exception XMLDBException
     */
    public String[] listXMLObjects() throws XMLDBException;

    /**
     * Creates a new collection level XMLObject using the provided configuration.
     * The XMLObject will be added to the collection using the provided name as
     * a key.
     *
     * @param configuration The XML configuration to use
     * @exception XMLDBException
     */
    public void createXMLObject(Document configuration) throws XMLDBException;

    /**
     * Drops a collection level XMLObject from the collection.
     *
     * @param name The name of the XMLObject to drop.
     * @exception XMLDBException
     */
    public void dropXMLObject(String name)
            throws XMLDBException;
}

