/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: DatabaseChangeObserver.java 564833 2007-08-11 04:25:23Z vgritsenko $
 */

package org.apache.xindice.core;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.core.data.Key;
import org.apache.xindice.core.data.Record;
import org.apache.xindice.util.Configuration;
import org.w3c.dom.Document;

/**
 * DatabaseChangeObserver is going to handle changes to collections by
 * flushing the apporate collection Filer.
 *
 * This current version we are only going to worry about document changes and
 * their Filer storage are kept updated.
 *
 * @author <a href="mailto:toddbyrne@gmail.com">Todd Byrne</a>
 * @version $Revision: 564833 $, $Date: 2007-08-11 00:25:23 -0400 (Sat, 11 Aug 2007) $
 */
public class DatabaseChangeObserver extends DBObserver {

    private static final Log log = LogFactory.getLog(DatabaseChangeObserver.class);

    /** One minute flush period **/
    private static final long FLUSH_PERIOD = 1000 * 60;

    // by using a hashtable instead of a queue collections only get added once.
    private final Hashtable collectionDirtyFlags;
    private Timer flushTimer;

    public DatabaseChangeObserver() {
        collectionDirtyFlags = new Hashtable();
        flushTimer = new Timer(true);
        flushTimer.schedule(new FlushTask(), FLUSH_PERIOD, FLUSH_PERIOD);
    }

    /**
     * Queues the collection to be flushed by the FlushTask synchronized
     * by the hashtable
     * @param col Collection
     * @param key Key
     * @throws DBException
     */
    public void dropDocument(Collection col, Key key) throws DBException {
        collectionDirtyFlags.put(col,Boolean.TRUE);
    }

    /**
     *
     * @param col Collection
     * @param key Key
     * @param document Document
     * @param create boolean
     * @throws DBException
     */
    public void putDocument(Collection col, Key key, Document document, boolean create) throws DBException {
        collectionDirtyFlags.put(col,Boolean.TRUE);
    }

    /**
     *
     * @param col Collection
     * @throws DBException
     */
    public void createCollection(Collection col) throws DBException {
    }

    /**
     *
     * @param col Collection
     * @param record Record
     * @param document Document
     * @throws DBException
     */
    public void loadDocument(Collection col, Record record, Document document) throws DBException {
    }

    /**
     *
     * @param col Collection
     * @throws DBException
     */
    public void dropCollection(Collection col) throws DBException {
    }

    /**
     *
     * @param db Database
     * @param cfg Configuration
     */
    public void flushDatabaseConfig(Database db, Configuration cfg) {
    }

    /**
     *
     * @param col Collection
     * @param cfg Configuration
     */
    public void setCollectionConfig(Collection col, Configuration cfg) {
    }

    /**
     *
     * @param db Database
     * @param collections Map
     * @param cfg Configuration
     */
    public void setDatabaseConfig(Database db, Map collections, Configuration cfg) {
    }

    /**
     * FlushTask handles the work of
     */
    private class FlushTask extends TimerTask {
        public FlushTask() {
        }

        public void run() {
            synchronized (collectionDirtyFlags) {
                int flushCount = 0;
                Iterator collections = collectionDirtyFlags.entrySet().iterator();
                while (collections.hasNext()) {
                    Map.Entry e = (Map.Entry) collections.next();
                    Collection c = (Collection) e.getKey();
                    try {
                        // don't need to flush indexes because everytime a document
                        // gets added the indexes get flushed.
                        c.getFiler().flush();
                        flushCount ++;
                    } catch (DBException ex) {
                        log.error("Couldn't flush collection: " + c.getName(), ex);
                    }
                }

                if (flushCount > 0) {
                    log.info("Successfully flushed " + flushCount + " collections out of " +
                             collectionDirtyFlags.size());
                }

                collectionDirtyFlags.clear();
            }
        }
    }

}
