/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
*/
package org.apache.xindice.core;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/**
 * Going to handle the JVM Shutdown hook and insuring clean shutdown
 * of all registered Database instances.
 *
 * <a href="mailto:toddbyrne@gmail.com">Todd Byrne</a>
 */
public class DatabaseShutdownHandler implements Runnable {

    private final Set databases;


    public DatabaseShutdownHandler() {
        databases = new HashSet();
        Runtime.getRuntime().addShutdownHook(new Thread(this));
    }

    /**
     * @param db database to register
     */
    public void registerDatabase(Database db) {
        synchronized (databases) {
            databases.add(db);
        }
    }

    /**
     * @param db removed the database from the hashtable
     */
    public void removeDatabase(Database db) {
        synchronized (databases) {
            databases.remove(db);
        }
    }

    /**
     * Cleans up any unclosed databases before the JVM exits.
     */
    public void run() {
        synchronized (databases) {
            Iterator dbs = databases.iterator();
            while (dbs.hasNext()) {
                Database db = (Database) dbs.next();

                // synch again on the db object just incase a close operation
                // is happening on a different thread
                synchronized (db) {
                    // call close but don't bother removing the db from the database
                    // table
                    try {
                        db.close(false);
                        System.out.println("SHUTDOWN Dirty close on: " + db.getName());
                    } catch (DBException e) {
                        System.out.println("SHUTDOWN Failed to close: " + db.getName() + ": " + e);
                    }
                }
            }

            databases.clear();
        }
    }
}
