/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: Entry.java 594588 2007-11-13 17:21:25Z vgritsenko $
 */

package org.apache.xindice.core.data;

import org.w3c.dom.Document;
import org.apache.xindice.xml.XMLSerializable;

import java.util.Map;
import java.util.HashMap;
import java.util.Collections;

/**
 * Entry is a high-level representation of a database record, that includes
 * its internal meta information.
 *
 * @version $Revision: 594588 $, $Date: 2007-11-13 12:21:25 -0500 (Tue, 13 Nov 2007) $
 */
public class Entry {

    public static final byte UNKNOWN  = 0;
    public static final byte DOCUMENT = 1;
    public static final byte BINARY   = 2;
    public static final byte OBJECT   = 3;

    public static final String CREATED  = "created";
    public static final String MODIFIED = "modified";

    private final byte type;
    private final Key key;
    private final Object value;
    private final Map meta;


    public Entry(byte type, Key key, Object value, Map meta) {
        this.type = type;
        this.key = key;
        this.value = value;
        this.meta = meta;
    }

    public Entry(Key key, Document value, Map meta) {
        this(DOCUMENT, key, value, meta);
    }

    public Entry(Key key, byte[] value, Map meta) {
        this(BINARY, key, value, meta);
    }

    public Entry(Key key, XMLSerializable value, Map meta) {
        this(OBJECT, key, value, meta);
    }

    public Entry(Key key, Map meta) {
        this(UNKNOWN, key, null, meta);
    }


    public byte getEntryType() {
        return type;
    }

    public Key getKey() {
        return key;
    }

    public Object getValue() {
        return value;
    }

    public Map getMeta() {
        return meta;
    }

    public long getCreationTime() {
        Long date = (Long) meta.get(CREATED);
        return date != null ? date.longValue() : 0;
    }

    public long getModificationTime() {
        Long date = (Long) meta.get(MODIFIED);
        return date != null ? date.longValue() : 0;
    }

    public static Map createMetaMap(Record record) {
        Map entryMeta = new HashMap(5);
        entryMeta.put(Entry.CREATED,  record.getMetaData(Record.CREATED));
        entryMeta.put(Entry.MODIFIED, record.getMetaData(Record.MODIFIED));
        return Collections.unmodifiableMap(entryMeta);
    }
}
