/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: Value.java 594315 2007-11-12 22:10:44Z vgritsenko $
 */

package org.apache.xindice.core.data;

import org.apache.xindice.util.XindiceRuntimeException;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;

/**
 * Value is the primary base class for all data storing objects.
 * The content window of value objects are immutable, but the
 * underlying byte array which was used for constructing the value might be not.
 *
 * @version $Revision: 594315 $, $Date: 2007-11-12 17:10:44 -0500 (Mon, 12 Nov 2007) $
 */
public class Value implements Comparable {

    private final byte[] data;
    private final int pos;
    private final int len;
    private int hash;

    /**
     * Usually there is no need to create a copy of the value, since it is
     * an immutable object. This constructor is mainly used to create
     * key objects.
     *
     * @param value the value object which data will be used to construct this
     *        value. 
     */
    public Value(Value value) {
        this.data = value.data;
        this.pos = value.pos;
        this.len = value.len;
    }

    public Value(byte[] data) {
        this.data = data;
        this.pos = 0;
        this.len = data.length;
    }

    public Value(byte[] data, int pos, int len) {
        if (pos >= data.length || pos < 0 || pos + len > data.length) {
            throw new ArrayIndexOutOfBoundsException("Value cannot be created");
        }

        this.data = data;
        this.pos = pos;
        this.len = len;
    }

    public Value(String data) {
        try {
            this.data = data.getBytes("utf-8");
            this.pos = 0;
            this.len = this.data.length;
        } catch (UnsupportedEncodingException e) {
            throw new XindiceRuntimeException("Java doesn't support UTF-8 encoding", e);
        }
    }

    /**
     * getLength retrieves the length of the data being stored by the Value.
     *
     * @return The Value length
     */
    public final int getLength() {
        return len;
    }

    /**
     * getData retrieves a copy of the data which is being stored
     * by this value as a byte array.
     *
     * <p>Data copying is performed in order to ensure immutability of the Value.
     * Avoid using this method if possible.
     *
     * @return The data
     */
    public final byte[] getData() {
        byte[] b = new byte[len];
        System.arraycopy(data, pos, b, 0, len);
        return b;
    }

    //
    // Data extraction
    //

    /**
     * Returns the byte at the specified index.
     *
     * @param index byte index
     * @return the byte at the specified index.
     * @throws ArrayIndexOutOfBoundsException if index is negative number or
     *         is not less that the length of Value data
     */
    public final byte byteAt(int index) {
        if (index < 0 || index >= len) {
            throw new ArrayIndexOutOfBoundsException(index);
        }
        return data[pos + index];
    }

    /**
     * Returns the short value at the specified index.
     *
     * @param index short index
     * @return the short at the specified index.
     * @throws ArrayIndexOutOfBoundsException if index is negative number or
     *         is not less that the length of the data array
     */
    public final short shortAt(int index) {
        return (short) ((data[index += pos] << 8) | data[index + 1]);
    }

    /**
     * Returns the int value at the specified index.
     *
     * @param index int index
     * @return the int at the specified index.
     * @throws ArrayIndexOutOfBoundsException if index is negative number or
     *         is not less that the length of the data array
     */
    public final int intAt(int index) {
        return (short) ((data[index += pos] << 24) | (data[index + 1] << 16) | (data[index + 2] << 8) | data[index + 3]);
    }

    /**
     * Get a value that is part of this value object.
     *
     * @param start beginning index
     * @param len length of the new value
     * @return Value object
     * @throws ArrayIndexOutOfBoundsException if start index is either negative
     *         or isn't less then length of original Value
     */
    public final Value valueAt(int start, int len) {
        return new Value(data, pos + start, len);
    }

    /**
     * Get a key that is part of this value object.
     *
     * @param start beginning index
     * @param len length of the new key
     * @return Key object
     * @throws ArrayIndexOutOfBoundsException if start index is either negative
     *         or isn't less then length of original Value
     */
    public final Key keyAt(int start, int len) {
        return new Key(data, pos + start, len);
    }

    //
    // I/O
    //

    /**
     * Return an InputStream for the value.
     *
     * @return An InputStream
     */
    public final InputStream getInputStream() {
        return new ByteArrayInputStream(data, pos, len);
    }

    /**
     * Stream the content of the value into an OutputStream.
     *
     * @param out the OutputStream
     * @throws IOException if write failed
     */
    public final void streamTo(OutputStream out) throws IOException {
        out.write(data, pos, len);
    }

    /**
     * Copy contents of the value into supplied byte array.
     *
     * @param tdata byte array for the value
     * @param tpos starting position
     */
    public final void copyTo(byte[] tdata, int tpos) {
        System.arraycopy(data, pos, tdata, tpos, len);
    }

    /**
     * Copy <code>len</code> bytes of value's content into supplied
     * byte array.
     *
     * @param tdata byte array for the value
     * @param tpos starting position
     * @param len count of bytes to copy
     */
    public final void copyTo(byte[] tdata, int tpos, int len) {
        System.arraycopy(data, pos, tdata, tpos, len);
    }

    //
    // Comparisons
    //

    public final boolean startsWith(Value value) {
        if (len < value.len) {
            return false;
        }

        byte[] ddata = value.data;
        int dpos = value.pos;

        for (int i = 0; i < value.len; i++) {
            if (data[i + pos] != ddata[i + dpos]) {
                return false;
            }
        }

        return true;
    }

    public final int compareTo(Value value) {
        byte[] ddata = value.data;
        int dpos = value.pos;
        int dlen = value.len;

        int stop = len > dlen ? dlen : len;

        for (int i = 0; i < stop; i++) {
            byte b1 = data[pos + i];
            byte b2 = ddata[dpos + i];

            if (b1 != b2) {
                // get unsigned value
                int s1 = ((int) b1) & 0xFF;
                int s2 = ((int) b2) & 0xFF;
                return s1 > s2 ? (i + 1) : -(i + 1);
            }
        }

        if (len == dlen) {
            return 0;
        } else {
            return len > dlen ? stop + 1 : -(stop + 1);
        }
    }

    public final int compareTo(Object obj) {
        if (obj instanceof Value) {
            return compareTo((Value) obj);
        }

        return compareTo(new Value(obj.toString()));
    }

    //
    // Object
    //

    public int hashCode() {
        // modeled after String.hashCode()
        if (hash == 0) {
            int tempHash = 0;
            for (int i = 0; i < len; i++) {
                tempHash = 31 * tempHash + data[pos + i];
            }
            this.hash = Math.abs(tempHash);
        }
        return hash;
    }

    public boolean equals(Value value) {
        return len == value.len && compareTo(value) == 0;
    }

    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }

        if (obj instanceof Value) {
            return equals((Value) obj);
        }

        return obj != null && equals(new Value(obj.toString()));
    }

    public final String toString() {
        try {
            return new String(data, pos, len, "utf-8");
        } catch (UnsupportedEncodingException e) {
            throw new XindiceRuntimeException("Java doesn't seem to support UTF-8!", e);
        }
    }
}
