/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: BTreeFiler.java 541516 2007-05-25 02:46:51Z vgritsenko $
 */

package org.apache.xindice.core.filer;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.core.DBException;
import org.apache.xindice.core.FaultCodes;
import org.apache.xindice.core.data.Key;
import org.apache.xindice.core.data.Record;
import org.apache.xindice.core.data.RecordSet;
import org.apache.xindice.core.data.Value;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

/**
 * BTreeFiler is a Filer implementation based on the BTree class.
 *
 * <br>
 * BTreeFiler has folowing configuration attributes:
 * <ul>
 * <li><strong>pagesize</strong>: Size of the page used by the filer.
 *     Default page size is 4096 bytes. This parameter can be set only
 *     before paged file is created. Once it is created, this parameter
 *     can not be changed.</li>
 * <li><strong>pagecount</strong>: Number of pages filer will be created
 *     with.</li>
 * <li><strong>maxkeysize</strong>: Maximum allowed size of the key.
 *     Default maximum key size is 256 bytes.</li>
 * <li><strong>max-descriptors</strong>: Defines maximum amount of
 *     simultaneously opened file descriptors this paged file can have.
 *     Several descriptors are needed to provide multithreaded access
 *     to the underlying file. Too large number will limit amount of
 *     collections you can open. Default value is 16
 *     (DEFAULT_DESCRIPTORS_MAX).</li>
 * </ul>
 *
 * @version $Revision: 541516 $, $Date: 2007-05-24 22:46:51 -0400 (Thu, 24 May 2007) $
 */
public class BTreeFiler extends BTree
                        implements Filer {

    private static final Log log = LogFactory.getLog(BTreeFiler.class);

    /**
     * Record page status
     */
    protected static final byte RECORD = 20;


    private BTreeFilerHeader fileHeader;


    public BTreeFiler() {
        super();
        fileHeader = (BTreeFilerHeader) getFileHeader();
    }

    public void setLocation(File root, String location) {
        setFile(new File(root, location + ".tbl"));
    }

    public String getName() {
        return "BTreeFiler";
    }

    public Record readRecord(Key key) throws DBException {
        return readRecord(key, false);
    }

    public Record readRecord(Key key, boolean metaOnly) throws DBException {
        if (key == null || key.getLength() == 0) {
            return null;
        }

        checkOpened();
        try {
            long pos = findValue(key);
            Page startPage = getPage(pos);
            Value v = metaOnly ? null : readValue(startPage);
            BTreeFilerPageHeader sph = (BTreeFilerPageHeader) startPage.getPageHeader();

            HashMap meta = new HashMap(2, 1.5F);
            meta.put(Record.CREATED, new Long(sph.getCreated()));
            meta.put(Record.MODIFIED, new Long(sph.getModified()));

            return new Record(key, v, meta);
        } catch (BTreeNotFoundException e) {
            if (log.isDebugEnabled()) {
                log.debug("Record '" + key + "' not found: " + e);
            }
        } catch (IOException e) {
            throw new FilerException(FaultCodes.DBE_CANNOT_READ,
                                     "Can't read record '" + key + "': " + e.getMessage(), e);
        }
        return null;
    }

    public Record writeRecord(Key key, Value value) throws DBException {
        if (key == null || key.getLength() == 0) {
            throw new FilerException(FaultCodes.DBE_CANNOT_CREATE, "Invalid key: null or empty");
        }
        if (value == null) {
            throw new FilerException(FaultCodes.DBE_CANNOT_CREATE, "Invalid value: null");
        }

        checkOpened();
        try {
            long pos = addKey(key);
            Page p = getPage(pos);

            BTreeFilerPageHeader ph = (BTreeFilerPageHeader) p.getPageHeader();
            long t = System.currentTimeMillis();
            if (ph.getStatus() == UNUSED) {
                ph.setCreated(t);
            }
            ph.setModified(t);
            ph.setStatus(RECORD);

            writeValue(p, value);
            flush();

            HashMap meta = new HashMap(3);
            meta.put(Record.CREATED, new Long(ph.getCreated()));
            meta.put(Record.MODIFIED, new Long(t));
            return new Record(key, value, meta);

        } catch (IOException e) {
            // FIXME: cleanup?
            throw new FilerException(FaultCodes.DBE_CANNOT_CREATE,
                                     "Can't write record '" + key + "': " + e.getMessage(), e);
        }
    }

    public boolean deleteRecord(Key key) throws DBException {
        if (key == null || key.getLength() == 0) {
            return false;
        }

        checkOpened();
        try {
            long pos = removeValue(key);
            Page p = getPage(pos);
            unlinkPages(p);

            fileHeader.decRecordCount();

            flush();
            return true;
        } catch (BTreeNotFoundException e) {
            if (log.isDebugEnabled()) {
                log.debug("Record '" + key + "' not found (" + e + ")");
            }
        } catch (IOException e) {
            throw new FilerException(FaultCodes.DBE_CANNOT_DROP,
                                     "Can't delete record '" + key + "': " + e.getMessage(), e);
        }

        return false;
    }

    public long getRecordCount() throws DBException {
        checkOpened();
        return fileHeader.getRecordCount();
    }

    public RecordSet getRecordSet() throws DBException {
        checkOpened();
        return new BTreeFilerRecordSet();
    }

    /**
     * BTreeFilerRecordSet
     */
    private class BTreeFilerRecordSet implements RecordSet, BTreeCallback {
        private List keys = new ArrayList();
        private Iterator iter;

        public BTreeFilerRecordSet() throws DBException {
            try {
                query(null, this);
                iter = keys.iterator();
            } catch (IOException e) {
                throw new FilerException(FaultCodes.GEN_CRITICAL_ERROR,
                                         "Error generating RecordSet", e);
            }
        }

        public synchronized boolean indexInfo(Value value, long pointer) {
            keys.add(new Key(value));
            return true;
        }

        public synchronized Key getNextKey() {
            return (Key) iter.next();
        }

        public synchronized Record getNextRecord() throws DBException {
            return readRecord((Key) iter.next(), false);
        }

        public synchronized Value getNextValue() throws DBException {
            return getNextRecord().getValue();
        }

        public synchronized boolean hasMoreRecords() {
            return iter.hasNext();
        }
    }

    ////////////////////////////////////////////////////////////////////

    protected FileHeader createFileHeader() {
        return new BTreeFilerHeader();
    }

    protected PageHeader createPageHeader() {
        return new BTreeFilerPageHeader();
    }

    /**
     * BTreeFilerHeader
     */
    private final class BTreeFilerHeader extends BTreeFileHeader {
        private long totalBytes;

        public BTreeFilerHeader() {
        }

        protected synchronized void read(RandomAccessFile raf) throws IOException {
            super.read(raf);
            totalBytes = raf.readLong();
        }

        protected synchronized void write(RandomAccessFile raf) throws IOException {
            super.write(raf);
            raf.writeLong(totalBytes);
        }

        /** The total number of bytes in use by the file */
        public synchronized void setTotalBytes(long totalBytes) {
            this.totalBytes = totalBytes;
            setDirty();
        }

        /** The total number of bytes in use by the file */
        public synchronized long getTotalBytes() {
            return totalBytes;
        }
    }

    /**
     * BTreeFilerPageHeader
     */
    private final class BTreeFilerPageHeader extends BTreePageHeader {
        private long created;
        private long modified;

        public BTreeFilerPageHeader() {
        }

        public BTreeFilerPageHeader(DataInput dis) throws IOException {
            super(dis);
        }

        public synchronized void read(DataInput dis) throws IOException {
            super.read(dis);

            if (getStatus() == UNUSED) {
                return;
            }

            created = dis.readLong();
            modified = dis.readLong();
        }

        public synchronized void write(DataOutput dos) throws IOException {
            super.write(dos);
            dos.writeLong(created);
            dos.writeLong(modified);
        }

        public synchronized void setRecordLen(int recordLen) {
            fileHeader.setTotalBytes((fileHeader.totalBytes - getRecordLen()) + recordLen);
            super.setRecordLen(recordLen);
        }

        /** UNIX-time when this record was created */
        public synchronized void setCreated(long created) {
            this.created = created;
            setDirty();
        }

        /** UNIX-time when this record was created */
        public synchronized long getCreated() {
            return created;
        }

        /** UNIX-time when this record was last modified */
        public synchronized void setModified(long modified) {
            this.modified = modified;
            setDirty();
        }

        /** UNIX-time when this record was last modified */
        public synchronized long getModified() {
            return modified;
        }
    }
}
