/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: FSFiler.java 541516 2007-05-25 02:46:51Z vgritsenko $
 */

package org.apache.xindice.core.filer;

import org.apache.xindice.core.DBException;
import org.apache.xindice.core.FaultCodes;
import org.apache.xindice.core.data.Key;
import org.apache.xindice.core.data.Record;
import org.apache.xindice.core.data.RecordSet;
import org.apache.xindice.core.data.Value;
import org.apache.xindice.util.Configuration;
import org.apache.xindice.util.FileCache;
import org.apache.xindice.util.LockManager;
import org.apache.xindice.util.SimpleConfigurable;
import org.apache.xindice.util.XindiceException;

import java.io.File;
import java.io.FileFilter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.StringTokenizer;

/**
 * FSFiler allows you to use existing file systems withing Xindice.
 *
 * @version $Revision: 541516 $, $Date: 2007-05-24 22:46:51 -0400 (Thu, 24 May 2007) $
 */
public final class FSFiler extends SimpleConfigurable implements Filer {

    // private static final Log log = LogFactory.getLog(FSFiler.class);

    private static final String LOCATION = "location";
    private static final String EXT = "ext";
    private static final String READONLY = "readonly";

    private FileCache cache = new FileCache();
    private LockManager locks = new LockManager(16);

    private Set extensions;

    private String location;
    private File dir;
    private boolean opened;
    private boolean readOnly;


    public FSFiler() {
    }

    public String getName() {
        return "FSFiler";
    }

    public void setLocation(File root, String location) {
    }

    public void setConfig(Configuration config) throws XindiceException {
        super.setConfig(config);
        location = config.getAttribute(LOCATION);
        readOnly = config.getBooleanAttribute(READONLY, readOnly);
        String exts = config.getAttribute(EXT);
        if (exts != null && exts.trim().length() > 0) {
            extensions = new HashSet();
            StringTokenizer st = new StringTokenizer(exts);
            while (st.hasMoreTokens()) {
                extensions.add(st.nextToken());
            }
        }
        dir = new File(location);
        opened = false;
    }

    private void checkOpened() throws DBException {
        if (!opened) {
            throw new FilerException(FaultCodes.COL_COLLECTION_CLOSED,
                                     "Filer is closed");
        }
    }

    private void checkReadOnly() throws DBException {
        if (readOnly) {
            throw new FilerException(FaultCodes.COL_COLLECTION_READ_ONLY,
                                     "Filer is read-only");
        }
    }

    public boolean close() {
        opened = false;
        return true;
    }

    public boolean open() {
        opened = (dir.exists() && dir.isDirectory());
        return opened;
    }

    public boolean drop() {
        opened = false;
        return true;
    }

    public boolean isOpened() {
        return opened;
    }

    public boolean exists() {
        return dir.exists();
    }

    public boolean create() {
        if (!dir.exists()) {
            return dir.mkdirs();
        } else {
            return true;
        }
    }

    public void flush() {
    }

    public Record readRecord(Key key) throws DBException {
        return readRecord(key, false);
    }

    public Record readRecord(Key key, boolean metaOnly) throws DBException {
        if (key == null || key.getLength() == 0) {
            return null;
        }

        checkOpened();

        String fname = key.toString();
        if (!isExtensionValid(fname)) {
            return null;
        }

        File file = new File(dir, fname);
        try {
            locks.acquireSharedLock(file);

            HashMap meta = new HashMap(2, 1.5F);
            meta.put(Record.MODIFIED, new Long(file.lastModified()));

            if (metaOnly && file.exists()) {
                return new Record(key, null, meta);
            } else {
                byte[] valueData = cache.getFile(file);
                if (valueData != null) {
                    return new Record(key, new Value(valueData), meta);
                }
            }
        } catch (IOException e) {
            throw new FilerException(FaultCodes.DBE_CANNOT_READ,
                                     "Can't read record '" + key + "': " + e.getMessage(), e);
        } finally {
            locks.releaseSharedLock(file);
        }
        return null;
    }

    public Record writeRecord(Key key, Value value) throws DBException {
        if (key == null || key.getLength() == 0) {
            throw new FilerException(FaultCodes.DBE_CANNOT_CREATE, "Invalid key: '" + key + "'");
        }
        if (value == null) {
            throw new FilerException(FaultCodes.DBE_CANNOT_CREATE, "Invalid null value");
        }

        checkOpened();
        checkReadOnly();

        String fname = key.toString();
        if (!isExtensionValid(fname)) {
            throw new FilerException(FaultCodes.DBE_CANNOT_CREATE, "Invalid extention");
        }

        File file = new File(dir, fname);
        try {
            locks.acquireExclusiveLock(file);
            FileOutputStream fos = new FileOutputStream(file);
            value.streamTo(fos);
            fos.close();

            HashMap meta = new HashMap(2);
            meta.put(Record.MODIFIED, new Long(file.lastModified()));

            return new Record(key, value, meta);
        } catch (IOException e) {
            throw new FilerException(FaultCodes.DBE_CANNOT_CREATE,
                                     "Can't write record '" + key + "': " + e.getMessage(), e);
        } finally {
            locks.releaseExclusiveLock(file);
        }
    }

    public boolean deleteRecord(Key key) throws DBException {
        if (key == null || key.getLength() == 0) {
            return false;
        }
        checkOpened();
        checkReadOnly();

        String fname = key.toString();
        if (!isExtensionValid(fname)) {
            return false;
        }

        File file = new File(dir, fname);
        try {
            locks.acquireExclusiveLock(file);
            // TODO: Should Exception (SecurityException) be catched here or not?
            return file.delete();
        } finally {
            locks.releaseExclusiveLock(file);
        }
    }

    public long getRecordCount() throws DBException {
        checkOpened();

        File[] files = dir.listFiles(new FileFilter() {
            public boolean accept(File file) {
                return file.isFile() && isExtensionValid(file.getName());
            }
        });
        return files.length;
    }

    public RecordSet getRecordSet() throws DBException {
        checkOpened();
        return new FSRecordSet();
    }

    private boolean isExtensionValid(String fname) {
        if (extensions != null) {
            int idx = fname.lastIndexOf('.');
            if (idx == -1) {
                return false;
            }
            String ext = fname.substring(idx + 1);
            if (!extensions.contains(ext)) {
                return false;
            }
        }
        return true;
    }

    /**
     * FSRecordSet
     */

    private class FSRecordSet implements RecordSet {
        public File[] files;
        public int pos = 0;

        public FSRecordSet() {
            files = dir.listFiles(new FileFilter() {
                public boolean accept(File file) {
                    return file.isFile() && isExtensionValid(file.getName());
                }
            });
        }

        public synchronized boolean hasMoreRecords() {
            return pos < files.length;
        }

        public synchronized Record getNextRecord() throws DBException {
            File file = files[pos++];
            return readRecord(new Key(file.getName()), false);
        }

        public synchronized Value getNextValue() throws DBException {
            return getNextRecord().getValue();
        }

        public synchronized Key getNextKey() {
            return new Key(files[pos++].getName());
        }
    }
}
