/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: Paged.java 571938 2007-09-02 10:14:13Z vgritsenko $
 */

package org.apache.xindice.core.filer;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.core.DBException;
import org.apache.xindice.core.DBObject;
import org.apache.xindice.core.FaultCodes;
import org.apache.xindice.core.data.Key;
import org.apache.xindice.core.data.Value;
import org.apache.xindice.util.Configurable;
import org.apache.xindice.util.Configuration;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInput;
import java.io.DataInputStream;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.lang.ref.WeakReference;
import java.util.Collection;
import java.util.EmptyStackException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Stack;
import java.util.WeakHashMap;

/**
 * Paged is a paged file implementation that is foundation for both the
 * BTree class and the HashFiler. It provides flexible paged I/O and
 * page caching functionality.
 *
 * <br>
 * Paged has folowing configuration attributes:
 * <ul>
 * <li><strong>pagesize</strong>: Size of the page used by the paged file.
 *     Default page size is 4096 bytes. This parameter can be set only
 *     before paged file is created. Once it is created, this parameter
 *     can not be changed.</li>
 * <li><strong>pagecount</strong>: Number of pages filer will be created
 *     with.</li>
 * <li><strong>maxkeysize</strong>: Maximum allowed size of the key.
 *     Default maximum key size is 256 bytes.</li>
 * <li><strong>max-descriptors</strong>: Defines maximum amount of
 *     simultaneously opened file descriptors this paged file can have.
 *     Several descriptors are needed to provide multithreaded access
 *     to the underlying file. Too large number will limit amount of
 *     collections you can open. Default value is 16
 *     (DEFAULT_DESCRIPTORS_MAX).</li>
 * </ul>
 *
 * <br>FIXME: Currently it seems that maxkeysize is not used anywhere.
 * <br>TODO: Introduce Paged interface, implementations.
 *
 * @version $Revision: 571938 $, $Date: 2007-09-02 06:14:13 -0400 (Sun, 02 Sep 2007) $
 */
public abstract class Paged implements DBObject, Configurable {

    private static final Log log = LogFactory.getLog(Paged.class);

    /**
     * The maximum number of pages that will be held in the dirty cache.
     * Once number reaches the limit, pages are flushed to disk.
     */
    private static final int MAX_DIRTY_SIZE = 128;

    /**
     * Name of the configuration attribute "pagesize"
     */
    protected static final String CONFIG_PAGESIZE = "pagesize";

    /**
     * Name of the configuration attribute "pagecount"
     */
    protected static final String CONFIG_PAGECOUNT = "pagecount";

    /**
     * Name of the configuration attribute "maxkeysize"
     */
    protected static final String CONFIG_KEYSIZE_MAX = "maxkeysize";

    /**
     * Name of the configuration attribute "max-descriptors"
     */
    protected static final String CONFIG_DESCRIPTORS_MAX = "max-descriptors";

    /**
     * Default value of the "pagesize".
     */
    private static final int DEFAULT_PAGESIZE = 4096;

    /**
     * Default value of the "pagecount".
     */
    private static final int DEFAULT_PAGECOUNT = 1024;

    /**
     * File header size
     */
    private static final int FILE_HEADER_SIZE = 4096;

    /**
     * Default value of the "maxkeysize".
     */
    private static final int DEFAULT_KEYSIZE_MAX = 256;

    /**
     * Default value of the maximum number of open random access files paged
     * can have. This number balances resources utilization and parallelism of
     * access to the paged file.
     */
    private static final int DEFAULT_DESCRIPTORS_MAX = 16;


    /**
     * Unused page status
     */
    protected static final byte UNUSED = 0;

    /**
     * Overflow page status
     */
    protected static final byte OVERFLOW = 126;

    /**
     * Deleted page status
     */
    protected static final byte DELETED = 127;

    /**
     * Page ID of non-existent page
     */
    protected static final int NO_PAGE = -1;


    /**
     * Configuration of this paged instance
     */
    private Configuration config;

    /**
     * Map of pages in use. Guarantees that page with same number will be loaded
     * into memory just once, allowing to synchronize on page objects to guarantee
     * no two threads are writing into same page at once.
     *
     * <p>Map contains weak references to the Page objects, keys are pages themselves.
     * Access is synchronized by the {@link #pagesLock}.
     */
    private final Map pages = new WeakHashMap();

    /**
     * Lock for synchronizing access to the {@link #pages} map.
     */
    private final Object pagesLock = new Object();

    /**
     * Cache of modified pages waiting to be written out.
     * Access is synchronized by the {@link #dirtyLock}.
     */
    private Map dirty = new HashMap();

    /**
     * Lock for synchronizing access to the {@link #dirty} map.
     */
    private final Object dirtyLock = new Object();

    /**
     * Random access file descriptors cache.
     * Access to it and to {@link #descriptorsCount} is synchronized by itself.
     */
    private final Stack descriptors = new Stack();

    /**
     * The number of random access file objects that exist, either in the
     * cache {@link #descriptors}, or currently in use.
     */
    private int descriptorsCount;

    /**
     * The maximum number of random access file objects that can be opened
     * by this paged instance.
     */
    private int descriptorsMax;

    /**
     * Whether the file is opened or not.
     */
    private boolean opened;

    /**
     * The underlying file where the Paged object stores its pages.
     */
    private File file;

    /**
     * Header of this Paged
     */
    private final FileHeader fileHeader;


    public Paged() {
        descriptorsMax = DEFAULT_DESCRIPTORS_MAX;
        fileHeader = createFileHeader();
    }

    public Paged(File file) {
        this();
        setFile(file);
    }

    public void setConfig(Configuration config) {
        this.config = config;
        // Read paged config
        fileHeader.setPageSize(config.getIntAttribute(CONFIG_PAGESIZE,
                                                      fileHeader.getPageSize()));
        fileHeader.setPageCount(config.getLongAttribute(CONFIG_PAGECOUNT,
                                                        fileHeader.getPageCount()));
        fileHeader.setMaxKeySize(config.getShortAttribute(CONFIG_KEYSIZE_MAX,
                                                          fileHeader.getMaxKeySize()));
        descriptorsMax = getConfig().getIntAttribute(CONFIG_DESCRIPTORS_MAX,
                                                     descriptorsMax);
    }

    public Configuration getConfig() {
        return config;
    }

    /**
     * setFile sets the file object for this Paged.
     *
     * @param file The File
     */
    protected final void setFile(final File file) {
        this.file = file;
    }

    /**
     * getFile returns the file object for this Paged.
     *
     * @return The File
     */
    protected final File getFile() {
        return file;
    }

    /**
     * Obtain RandomAccessFile ('descriptor') object out of the pool.
     * If no descriptors available, and maximum amount already allocated,
     * the call will block.
     */
    protected final RandomAccessFile getDescriptor() throws IOException {
        synchronized (descriptors) {
            // If there are descriptors in the cache return one.
            if (!descriptors.empty()) {
                return (RandomAccessFile) descriptors.pop();
            }
            // Otherwise we need to get one some other way.

            // First try to create a new one if there's room
            if (descriptorsCount < descriptorsMax) {
                descriptorsCount++;
                return new RandomAccessFile(file, "rw");
            }

            // Otherwise we have to wait for one to be released by another thread.
            while (true) {
                try {
                    descriptors.wait();
                    return (RandomAccessFile) descriptors.pop();
                } catch (InterruptedException e) {
                    // Ignore, and continue to wait
                } catch (EmptyStackException e) {
                    // Ignore, and continue to wait
                }
            }
        }
    }

    /**
     * Puts a RandomAccessFile ('descriptor') back into the descriptor pool.
     */
    protected final void putDescriptor(RandomAccessFile raf) {
        if (raf != null) {
            synchronized (descriptors) {
                descriptors.push(raf);
                descriptors.notify();
            }
        }
    }

    /**
     * Closes a RandomAccessFile ('descriptor') and removes it from the pool.
     */
    protected final void closeDescriptor(RandomAccessFile raf) {
        if (raf != null) {
            try {
                raf.close();
            } catch (IOException e) {
                // Ignore close exception
            }

            // Synchronization is necessary as decrement operation is not atomic
            synchronized (descriptors) {
                descriptorsCount --;
            }
        }
    }

    /**
     * getPage returns the page specified by pageNum.
     *
     * @param pageNum The Page number
     * @return The requested Page
     * @throws IOException if an Exception occurs
     */
    protected final Page getPage(long pageNum) throws IOException {
        final PageKey k = new PageKey(pageNum);
        Page p = null;
        synchronized (pagesLock) {
            // Check if page is already loaded in the page cache
            WeakReference ref = (WeakReference) pages.get(k);
            if (ref != null) {
                p = (Page) ref.get();
                // Fall through to p.read(). Even if page present in the pages
                // map, it still has to be read - it could be that it was just
                // added to the map but read() was not called yet.
            }

            // If not found, create it and add it to the pages cache
            if (p == null) {
                p = new Page(pageNum);
                pages.put(p, new WeakReference(p));
            }
        }

        // Load the page from disk if necessary
        p.read();
        return p;
    }

    /**
     * readValue reads the multi-Paged Value starting at the specified
     * Page.
     *
     * @param page The starting Page
     * @return The Value
     * @throws IOException if an Exception occurs
     */
    protected final Value readValue(Page page) throws IOException {
        final PageHeader sph = page.getPageHeader();
        ByteArrayOutputStream bos = new ByteArrayOutputStream(sph.getRecordLen());

        // Loop until we've read all the pages into memory.
        Page p = page;
        while (true) {
            PageHeader ph = p.getPageHeader();

            // Add the contents of the page onto the stream
            p.streamTo(bos);

            // Continue following the list of pages until we get to the end.
            long nextPage = ph.getNextPage();
            if (nextPage == NO_PAGE) {
                break;
            }
            p = getPage(nextPage);
        }

        // Return a Value with the collected contents of all pages.
        return new Value(bos.toByteArray());
    }

    /**
     * readValue reads the multi-Paged Value starting at the specified
     * page number.
     *
     * @param page The starting page number
     * @return The Value
     * @throws IOException if an Exception occurs
     */
    protected final Value readValue(long page) throws IOException {
        return readValue(getPage(page));
    }

    /**
     * writeValue writes the multi-Paged Value starting at the specified
     * Page.
     *
     * @param page The starting Page
     * @param value The Value to write
     * @throws IOException if an Exception occurs
     */
    protected final void writeValue(Page page, Value value) throws IOException {
        if (value == null) {
            throw new IOException("Can't write a null value");
        }

        InputStream is = value.getInputStream();

        // Write as much as we can onto the primary page.
        PageHeader hdr = page.getPageHeader();
        hdr.setRecordLen(value.getLength());
        page.streamFrom(is);

        // Write out the rest of the value onto any needed overflow pages
        while (is.available() > 0) {
            Page lpage = page;
            PageHeader lhdr = hdr;

            // Find an overflow page to use
            long np = lhdr.getNextPage();
            if (np != NO_PAGE) {
                // Use an existing page.
                page = getPage(np);
            } else {
                // Create a new overflow page
                page = getFreePage();
                lhdr.setNextPage(page.getPageNum());
            }

            // Mark the page as an overflow page.
            hdr = page.getPageHeader();
            hdr.setStatus(OVERFLOW);

            // Write some more of the value to the overflow page.
            page.streamFrom(is);
            lpage.write();
        }

        // Cleanup any unused overflow pages. i.e. the value is smaller then the
        // last time it was written.
        long np = hdr.getNextPage();
        if (np != NO_PAGE) {
            unlinkPages(np);
        }

        hdr.setNextPage(NO_PAGE);
        page.write();
    }

    /**
     * writeValue writes the multi-Paged Value starting at the specified
     * page number.
     *
     * @param page The starting page number
     * @param value The Value to write
     * @throws IOException if an Exception occurs
     */
    protected final void writeValue(long page, Value value) throws IOException {
        writeValue(getPage(page), value);
    }

    /**
     * unlinkPages unlinks a set of pages starting at the specified Page.
     *
     * @param page The starting Page to unlink
     * @throws IOException if an Exception occurs
     */
    protected void unlinkPages(Page page) throws IOException {
        // Add any overflow pages to the list of free pages.
        // Get the first and last page in the chain.
        long firstPage = page.pageNum;
        while (page.header.nextPage != NO_PAGE) {
            page = getPage(page.header.nextPage);
        }
        long lastPage = page.pageNum;

        // Free the chain
        synchronized (fileHeader) {
            // If there are already some free pages, add the start of the chain
            // to the list of free pages.
            if (fileHeader.lastFreePage != NO_PAGE) {
                Page p = getPage(fileHeader.lastFreePage);
                p.header.setNextPage(firstPage);
                p.write();
            }

            // Otherwise set the chain as the list of free pages.
            if (fileHeader.firstFreePage == NO_PAGE) {
                fileHeader.setFirstFreePage(firstPage);
            }

            // Add a reference to the end of the chain.
            fileHeader.setLastFreePage(lastPage);
        }
    }

    /**
     * unlinkPages unlinks a set of pages starting at the specified
     * page number.
     *
     * @param pageNum The starting page number to unlink
     * @throws IOException if an Exception occurs
     */
    protected final void unlinkPages(long pageNum) throws IOException {
        unlinkPages(getPage(pageNum));
    }

    /**
     * getFreePage returns the first free Page from secondary storage.
     * If no Pages are available, the file is grown as appropriate.
     *
     * @return The next free Page
     * @throws IOException if an Exception occurs
     */
    protected final Page getFreePage() throws IOException {
        Page p = null;

        // Synchronize read and write to the fileHeader.firstFreePage
        synchronized (fileHeader) {
            if (fileHeader.firstFreePage != NO_PAGE) {
                // Steal a deleted page
                p = getPage(fileHeader.firstFreePage);
                fileHeader.setFirstFreePage(p.getPageHeader().nextPage);
                if (fileHeader.firstFreePage == NO_PAGE) {
                    fileHeader.setLastFreePage(NO_PAGE);
                }
            }
        }

        if (p == null) {
            // No deleted pages, grow the file
            p = getPage(fileHeader.incTotalCount());
        }

        // Initialize The Page Header (Cleanly)
        p.header.setNextPage(NO_PAGE);
        p.header.setStatus(UNUSED);
        return p;
    }

    /**
     * @throws DBException COL_COLLECTION_CLOSED if paged file is closed
     */
    protected final void checkOpened() throws DBException {
        if (!opened) {
            throw new FilerException(FaultCodes.COL_COLLECTION_CLOSED,
                                     "Filer is closed");
        }
    }

    /**
     * getFileHeader returns the FileHeader
     *
     * @return The FileHeader
     */
    protected FileHeader getFileHeader() {
        return fileHeader;
    }

    /**
     * @return True if this paged file exists
     */
    public boolean exists() {
        return file.exists();
    }

    public boolean create() throws DBException {
        try {
            createFile();
            fileHeader.write();
            flush();
            return true;
        } catch (Exception e) {
            throw new FilerException(FaultCodes.GEN_CRITICAL_ERROR,
                                     "Error creating " + file.getName(), e);
        }
    }

    private void createFile() throws IOException {
        RandomAccessFile raf = null;
        try {
            raf = getDescriptor();
            long o = fileHeader.headerSize + (fileHeader.pageCount + 1) * fileHeader.pageSize - 1;
            raf.seek(o);
            raf.write(0);
        }  finally {
            putDescriptor(raf);
        }
    }

    public synchronized boolean open() throws DBException {
        RandomAccessFile raf = null;
        try {
            if (exists()) {
                raf = getDescriptor();
                fileHeader.read();

                // This is the only property that can be changed after creation
                fileHeader.setMaxKeySize(config.getShortAttribute(CONFIG_KEYSIZE_MAX,
                                                                  fileHeader.getMaxKeySize()));

                opened = true;
            } else {
                opened = false;
            }
            return opened;
        } catch (Exception e) {
            throw new FilerException(FaultCodes.GEN_CRITICAL_ERROR,
                                     "Error opening " + file.getName(), e);
        } finally {
            putDescriptor(raf);
        }
    }

    public synchronized boolean close() throws DBException {
        if (isOpened()) {
            try {
                // First of all, mark as closed to prevent operations
                opened = false;
                flush();

                synchronized (descriptors) {
                    final int total = descriptorsCount;
                    // Close descriptors in cache
                    while (!descriptors.empty()) {
                        closeDescriptor((RandomAccessFile)descriptors.pop());
                    }
                    // Attempt to close descriptors in use. Max wait time = 0.5s * MAX_DESCRIPTORS
                    int n = descriptorsCount;
                    while (descriptorsCount > 0 && n > 0) {
                        descriptors.wait(500);
                        if (descriptors.isEmpty()) {
                            n--;
                        } else {
                            closeDescriptor((RandomAccessFile)descriptors.pop());
                        }
                    }
                    if (descriptorsCount > 0) {
                        log.warn(descriptorsCount + " out of " + total + " files were not closed during close.");
                    }
                }

                // clear cache
                synchronized (pagesLock) {
                    pages.clear();
                }
            } catch (Exception e) {
                // Failed to close, leave open
                opened = true;
                throw new FilerException(FaultCodes.GEN_CRITICAL_ERROR,
                                         "Error closing " + file.getName(), e);
            }
        }

        return true;
    }

    public boolean isOpened() {
        return opened;
    }

    public boolean drop() throws DBException {
        try {
            close();
            if (exists()) {
                return getFile().delete();
            } else {
                return true;
            }
        } catch (Exception e) {
            throw new FilerException(FaultCodes.COL_CANNOT_DROP,
                                     "Can't drop " + file.getName(), e);
        }
    }

    void addDirty(Page page) throws IOException {
        boolean flush;
        synchronized (dirtyLock) {
            dirty.put(page, page);
            flush = dirty.size() > MAX_DIRTY_SIZE;
        }

        if (flush) {
            // Too many dirty pages... flush them
            try {
                flush();
            } catch (Exception e) {
                throw new IOException(e.getMessage());
            }
        }
    }

    public void flush() throws DBException {
        // This method is not synchronized

        // Error flag/counter
        int error = 0;

        // Obtain collection of dirty pages
        Collection pages;
        synchronized (dirtyLock) {
           pages = dirty.values();
           dirty = new HashMap();
        }

        // Flush dirty pages
        Iterator i = pages.iterator();
        while (i.hasNext()) {
            Page p = (Page) i.next();
            try {
                p.flush();
            } catch (Exception e) {
                log.warn("Exception while flushing page " + p.pageNum, e);
                error++;
            }
        }

        // Flush header
        if (fileHeader.dirty) {
            try {
                fileHeader.write();
            } catch (Exception e) {
                log.warn("Exception while flushing file header", e);
                error++;
            }
        }

        if (error != 0) {
            throw new FilerException(FaultCodes.GEN_CRITICAL_ERROR,
                                     "Error performing flush! Failed to flush " + error + " pages!");
        }
    }


    /**
     * createFileHeader must be implemented by a Paged implementation
     * in order to create an appropriate subclass instance of a FileHeader.
     *
     * @return a new FileHeader
     */
    protected abstract FileHeader createFileHeader();

    /**
     * createPageHeader must be implemented by a Paged implementation
     * in order to create an appropriate subclass instance of a PageHeader.
     *
     * @return a new PageHeader
     */
    protected abstract PageHeader createPageHeader();


    // These are a bunch of utility methods for subclasses

    public static Value[] insertArrayValue(Value[] vals, Value val, int idx) {
        Value[] newVals = new Value[vals.length + 1];
        if (idx > 0) {
            System.arraycopy(vals, 0, newVals, 0, idx);
        }
        newVals[idx] = val;
        if (idx < vals.length) {
            System.arraycopy(vals, idx, newVals, idx + 1, vals.length - idx);
        }
        return newVals;
    }

    public static Value[] deleteArrayValue(Value[] vals, int idx) {
        Value[] newVals = new Value[vals.length - 1];
        if (idx > 0) {
            System.arraycopy(vals, 0, newVals, 0, idx);
        }
        if (idx < newVals.length) {
            System.arraycopy(vals, idx + 1, newVals, idx, newVals.length - idx);
        }
        return newVals;
    }

    public static long[] insertArrayLong(long[] vals, long val, int idx) {
        long[] newVals = new long[vals.length + 1];
        if (idx > 0) {
            System.arraycopy(vals, 0, newVals, 0, idx);
        }
        newVals[idx] = val;
        if (idx < vals.length) {
            System.arraycopy(vals, idx, newVals, idx + 1, vals.length - idx);
        }
        return newVals;
    }

    public static long[] deleteArrayLong(long[] vals, int idx) {
        long[] newVals = new long[vals.length - 1];
        if (idx > 0) {
            System.arraycopy(vals, 0, newVals, 0, idx);
        }
        if (idx < newVals.length) {
            System.arraycopy(vals, idx + 1, newVals, idx, newVals.length - idx);
        }
        return newVals;
    }

    public static int[] insertArrayInt(int[] vals, int val, int idx) {
        int[] newVals = new int[vals.length + 1];
        if (idx > 0) {
            System.arraycopy(vals, 0, newVals, 0, idx);
        }
        newVals[idx] = val;
        if (idx < vals.length) {
            System.arraycopy(vals, idx, newVals, idx + 1, vals.length - idx);
        }
        return newVals;
    }

    public static int[] deleteArrayInt(int[] vals, int idx) {
        int[] newVals = new int[vals.length - 1];
        if (idx > 0) {
            System.arraycopy(vals, 0, newVals, 0, idx);
        }
        if (idx < newVals.length) {
            System.arraycopy(vals, idx + 1, newVals, idx, newVals.length - idx);
        }
        return newVals;
    }

    public static short[] insertArrayShort(short[] vals, short val, int idx) {
        short[] newVals = new short[vals.length + 1];
        if (idx > 0) {
            System.arraycopy(vals, 0, newVals, 0, idx);
        }
        newVals[idx] = val;
        if (idx < vals.length) {
            System.arraycopy(vals, idx, newVals, idx + 1, vals.length - idx);
        }

        return newVals;
    }

    public static short[] deleteArrayShort(short[] vals, int idx) {
        short[] newVals = new short[vals.length - 1];
        if (idx > 0) {
            System.arraycopy(vals, 0, newVals, 0, idx);
        }
        if (idx < newVals.length) {
            System.arraycopy(vals, idx + 1, newVals, idx, newVals.length - idx);
        }

        return newVals;
    }


    /**
     * Paged file's header
     */
    protected abstract class FileHeader {
        private boolean dirty;
        private int workSize;

        private short headerSize;

        /**
         * Size of the page in bytes.
         */
        private int pageSize;

        /**
         * Number of pages initially allocated for the file.
         * Has a special (historical) meaning for HashFiler.
         */
        private long pageCount;

        /**
         * Number of pages used by the filer. Initially set to 0.
         * Has somewhat different (historical) meaning for HashFiler.
         */
        private long totalCount;

        private long firstFreePage = -1;
        private long lastFreePage = -1;
        private byte pageHeaderSize = 64;
        private short maxKeySize = DEFAULT_KEYSIZE_MAX;
        private long recordCount;


        public FileHeader() {
            this.pageSize = DEFAULT_PAGESIZE;
            this.pageCount = DEFAULT_PAGECOUNT;
            this.headerSize = (short) FILE_HEADER_SIZE;
            calculateWorkSize();
        }

        public synchronized final void read() throws IOException {
            RandomAccessFile raf = getDescriptor();
            try {
                raf.seek(0);
                read(raf);
                calculateWorkSize();
            } finally {
                putDescriptor(raf);
            }
        }

        protected synchronized void read(RandomAccessFile raf) throws IOException {
            headerSize = raf.readShort();
            pageSize = raf.readInt();
            pageCount = raf.readLong();
            totalCount = raf.readLong();
            firstFreePage = raf.readLong();
            lastFreePage = raf.readLong();
            pageHeaderSize = raf.readByte();
            maxKeySize = raf.readShort();
            recordCount = raf.readLong();
        }

        public synchronized final void write() throws IOException {
            if (dirty) {
                RandomAccessFile raf = getDescriptor();
                try {
                    raf.seek(0);
                    write(raf);
                    dirty = false;
                } finally {
                    putDescriptor(raf);
                }
            }
        }

        protected synchronized void write(RandomAccessFile raf) throws IOException {
            raf.writeShort(headerSize);
            raf.writeInt(pageSize);
            raf.writeLong(pageCount);
            raf.writeLong(totalCount);
            raf.writeLong(firstFreePage);
            raf.writeLong(lastFreePage);
            raf.writeByte(pageHeaderSize);
            raf.writeShort(maxKeySize);
            raf.writeLong(recordCount);
        }

        public synchronized final void setDirty() {
            dirty = true;
        }

        public synchronized final boolean isDirty() {
            return dirty;
        }

        /**
         * The size of the FileHeader. Usually 1 OS Page.
         * This method should be called only while initializing Paged, not during normal processing.
         */
        public synchronized final void setHeaderSize(short headerSize) {
            this.headerSize = headerSize;
            dirty = true;
        }

        /** The size of the FileHeader.  Usually 1 OS Page */
        public synchronized final short getHeaderSize() {
            return headerSize;
        }

        /**
         * The size of a page. Usually a multiple of a FS block.
         * This method should be called only while initializing Paged, not during normal processing.
         */
        public synchronized final void setPageSize(int pageSize) {
            this.pageSize = pageSize;
            calculateWorkSize();
            dirty = true;
        }

        /** The size of a page.  Usually a multiple of a FS block */
        public synchronized final int getPageSize() {
            return pageSize;
        }

        /**
         * The number of pages in primary/initial storage.
         * This method should be called only while initializing Paged, not during normal processing.
         */
        public synchronized final void setPageCount(long pageCount) {
            this.pageCount = pageCount;
            dirty = true;
        }

        /** The number of pages in primary storage */
        public synchronized final long getPageCount() {
            return pageCount;
        }

        /**
         * The number of used pages in the file.
         * This method should be called only while initializing Paged, not during normal processing.
         */
        public synchronized final void setTotalCount(long totalCount) {
            this.totalCount = totalCount;
            dirty = true;
        }

        public synchronized final long incTotalCount() {
            dirty = true;
            return this.totalCount++;
        }

        /** The number of used pages in the file */
        public synchronized final long getTotalCount() {
            return totalCount;
        }

        /** The first free page in unused secondary space */
        public synchronized final void setFirstFreePage(long firstFreePage) {
            this.firstFreePage = firstFreePage;
            dirty = true;
        }

        /** The first free page in unused secondary space */
        public synchronized final long getFirstFreePage() {
            return firstFreePage;
        }

        /** The last free page in unused secondary space */
        public synchronized final void setLastFreePage(long lastFreePage) {
            this.lastFreePage = lastFreePage;
            dirty = true;
        }

        /** The last free page in unused secondary space */
        public synchronized final long getLastFreePage() {
            return lastFreePage;
        }

        /**
         * Set the size of a page header.
         *
         * Normally, 64 is sufficient.
         */
        public synchronized final void setPageHeaderSize(byte pageHeaderSize) {
            this.pageHeaderSize = pageHeaderSize;
            calculateWorkSize();
            dirty = true;
        }

        /**
         * Get the size of a page header.
         */
        public synchronized final byte getPageHeaderSize() {
            return pageHeaderSize;
        }

        /**
         * Set the maximum number of bytes a key can be.
         *
         * Normally, 256 is good
         */
        public synchronized final void setMaxKeySize(short maxKeySize) {
            this.maxKeySize = maxKeySize;
            dirty = true;
        }

        /**
         * Get the maximum number of bytes a key can be.
         */
        public synchronized final short getMaxKeySize() {
            return maxKeySize;
        }

        /** Increment the number of records being managed by the file */
        public synchronized final void incRecordCount() {
            recordCount++;
            dirty = true;
        }

        /** Decrement the number of records being managed by the file */
        public synchronized final void decRecordCount() {
            recordCount--;
            dirty = true;
        }

        /** The number of records being managed by the file (not pages) */
        public synchronized final long getRecordCount() {
            return recordCount;
        }

        private synchronized void calculateWorkSize() {
            workSize = pageSize - pageHeaderSize;
        }

        public synchronized final int getWorkSize() {
            return workSize;
        }
    }

    /**
     * Paged file page's header
     */
    protected abstract static class PageHeader implements Streamable {
        private boolean dirty;
        private byte status = UNUSED;
        private short keyLen;
        private int keyHash;
        private int dataLen;
        private int recordLen;
        private long nextPage = NO_PAGE;

        public PageHeader() {
        }

        public PageHeader(DataInput dis) throws IOException {
            read(dis);
        }

        public synchronized void read(DataInput dis) throws IOException {
            status = dis.readByte();
            dirty = false;
            if (status == UNUSED) {
                return;
            }

            keyLen = dis.readShort();
            keyHash = dis.readInt();
            dataLen = dis.readInt();
            recordLen = dis.readInt();
            nextPage = dis.readLong();
        }

        public synchronized void write(DataOutput dos) throws IOException {
            dirty = false;
            dos.writeByte(status);
            dos.writeShort(keyLen);
            dos.writeInt(keyHash);
            dos.writeInt(dataLen);
            dos.writeInt(recordLen);
            dos.writeLong(nextPage);
        }

        public synchronized final boolean isDirty() {
            return dirty;
        }

        public synchronized final void setDirty() {
            dirty = true;
        }

        /** The status of this page (UNUSED, RECORD, DELETED, etc...) */
        public synchronized final void setStatus(byte status) {
            this.status = status;
            dirty = true;
        }

        /** The status of this page (UNUSED, RECORD, DELETED, etc...) */
        public synchronized final byte getStatus() {
            return status;
        }

        public synchronized final void setKey(Key key) {
            // setKey WIPES OUT the Page data
            setRecordLen(0);
            dataLen = 0;
            keyHash = key.hashCode();
            keyLen = (short) key.getLength();
            dirty = true;
        }

        /** The length of the Key */
        public synchronized final void setKeyLen(short keyLen) {
            this.keyLen = keyLen;
            dirty = true;
        }

        /** The length of the Key */
        public synchronized final short getKeyLen() {
            return keyLen;
        }

        /** The hashed value of the Key for quick comparisons */
        public synchronized final void setKeyHash(int keyHash) {
            this.keyHash = keyHash;
            dirty = true;
        }

        /** The hashed value of the Key for quick comparisons */
        public synchronized final int getKeyHash() {
            return keyHash;
        }

        /** The length of the Data */
        public synchronized final void setDataLen(int dataLen) {
            this.dataLen = dataLen;
            dirty = true;
        }

        /** The length of the Data */
        public synchronized final int getDataLen() {
            return dataLen;
        }

        /** The length of the Record's value */
        public synchronized void setRecordLen(int recordLen) {
            this.recordLen = recordLen;
            dirty = true;
        }

        /** The length of the Record's value */
        public synchronized final int getRecordLen() {
            return recordLen;
        }

        /** The next page for this Record (if overflowed) */
        public synchronized final void setNextPage(long nextPage) {
            this.nextPage = nextPage;
            dirty = true;
        }

        /** The next page for this Record (if overflowed) */
        public synchronized final long getNextPage() {
            return nextPage;
        }
    }

    /**
     * The object wrapping page number.
     */
    protected static class PageKey implements Comparable {

        /**
         * This page number
         */
        protected final long pageNum;


        public PageKey(long pageNum) {
            this.pageNum = pageNum;
        }

        // No synchronization - pageNum is final
        public long getPageNum() {
            return pageNum;
        }

        // No synchronization: pageNum is final.
        public int compareTo(Object o) {
            return (int) (pageNum - ((PageKey) o).pageNum);
        }

        /**
         * Return page hash code, which is hash code of its {@link #pageNum}.
         *
         * @return Page hash code
         */
        public int hashCode() {
            // Unroll new Long(pageNum).hashCode()
            return (int) (pageNum ^ (pageNum >> 32));
        }

        /**
         * Pages are equal if they are the same or have equal pageNum.
         *
         * @param obj Another page
         * @return true if pages are equal
         */
        public boolean equals(Object obj) {
            if (obj == this) {
                return true;
            }

            if (obj instanceof PageKey) {
                return pageNum == ((PageKey) obj).pageNum;
            }

            return false;
        }
    }

    /**
     * Paged file's page
     */
    protected final class Page extends PageKey {

        /**
         * The Header for this Page
         */
        private final PageHeader header;

        /**
         * The offset into the file that this page starts
         */
        private final long offset;

        /**
         * The data for this page. Null if page is not loaded.
         */
        private byte[] data;

        /**
         * The position (relative) of the Key in the data array
         */
        private int keyPos;

        /**
         * The position (relative) of the Data in the data array
         */
        private int dataPos;


        private Page(long pageNum) {
            super(pageNum);
            this.header = createPageHeader();
            this.offset = fileHeader.headerSize + (pageNum * fileHeader.pageSize);
        }

        /**
         * Reads a page into the memory, once. Subsequent calls are ignored.
         */
        public synchronized void read() throws IOException {
            if (data == null) {
                RandomAccessFile raf = null;
                try {
                    byte[] data = new byte[fileHeader.pageSize];
                    raf = getDescriptor();
                    raf.seek(this.offset);
                    raf.read(data);

                    // Read in the header
                    ByteArrayInputStream bis = new ByteArrayInputStream(data);
                    this.header.read(new DataInputStream(bis));

                    this.keyPos = fileHeader.pageHeaderSize;
                    this.dataPos = this.keyPos + this.header.keyLen;

                    // Successfully read all the data
                    this.data = data;
                } finally {
                    putDescriptor(raf);
                }
            }
        }

        /**
         * Writes out the header into the this.data, and adds a page to the set of
         * dirty pages.
         */
        public void write() throws IOException {
            // Write out the header into the this.data
            synchronized (this) {
                ByteArrayOutputStream bos = new ByteArrayOutputStream(fileHeader.getPageHeaderSize());
                header.write(new DataOutputStream(bos));
                byte[] b = bos.toByteArray();
                System.arraycopy(b, 0, data, 0, b.length);
            }

            // Add to the list of dirty pages
            Paged.this.addDirty(this);
        }

        /**
         * Flushes content of the dirty page into the file
         */
        public synchronized void flush() throws IOException {
            RandomAccessFile raf = null;
            try {
                raf = getDescriptor();
                if (this.offset >= raf.length()) {
                    // Grow the file
                    long o = fileHeader.headerSize + (fileHeader.totalCount * 3 / 2 + 1) * fileHeader.pageSize - 1;
                    raf.seek(o);
                    raf.writeByte(0);
                }
                raf.seek(this.offset);
                raf.write(this.data);
            } finally {
                putDescriptor(raf);
            }
        }

        // No synchronization - header is final
        public PageHeader getPageHeader() {
            return this.header;
        }

        public synchronized void setKey(Key key) {
            header.setKey(key);
            // Insert the key into the data array.
            key.copyTo(this.data, this.keyPos);

            // Set the start of data to skip over the key.
            this.dataPos = this.keyPos + header.keyLen;
        }

        public synchronized Key getKey() {
            if (header.keyLen == 0) {
                return null;
            }

            return new Key(this.data, this.keyPos, header.keyLen);
        }

        public synchronized void streamTo(OutputStream os) throws IOException {
            if (header.dataLen > 0) {
                os.write(this.data, this.dataPos, header.dataLen);
            }
        }

        public synchronized void streamFrom(InputStream is) throws IOException {
            int avail = is.available();
            header.dataLen = fileHeader.workSize - header.keyLen;
            if (avail < header.dataLen) {
                header.dataLen = avail;
            }
            if (header.dataLen > 0) {
                is.read(this.data, this.keyPos + header.keyLen, header.dataLen);
            }
        }
    }
}
