/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: IndexerEventHandler.java 564424 2007-08-09 23:40:03Z natalia $
 */

package org.apache.xindice.core.indexer;

import org.apache.xindice.core.data.Key;
import org.apache.xindice.core.DBException;

/**
 * IndexerEventHandler represents an event handler for Indexer-specific events.
 * It is used by Indexers to manage adding and removing data.
 *
 * @see Indexer
 * @version $Revision: 564424 $, $Date: 2007-08-09 19:40:03 -0400 (Thu, 09 Aug 2007) $
 */
public interface IndexerEventHandler {
    /**
     * onDocumentAdded method is called after document is added and all its
     * elements are processed. This method will be called regardless of
     * Indexer patterns, even if no elements/attributes in that document
     * matches any of Indexer patterns.
     *
     * @param key Document key
     * @throws DBException
     */
    void onDocumentAdded(Key key) throws DBException;

    /**
     * onDocumentDeleted method is called after document is deleted and all its
     * elements are processed. This method will be called regardless of Indexer
     * patterns, even if no elements/attributes in that document matches any of
     * Indexer patterns.
     *
     * @param key Document key
     * @throws DBException
     */
    void onDocumentDeleted(Key key) throws DBException;

    /**
     * onNameAdded called for every element/attribute in the document being added
     * that matches any of the Indexer pattern.
     *
     * @param pattern Indexer pattern that matches the element/attribute
     * @param key     Document key
     * @param elemID  Element ID that matches the pattern
     * @param attrID  Attribute ID that matches the pattern
     * @throws DBException
     */
    void onNameAdded(IndexPattern pattern, Key key, short elemID, short attrID) throws DBException;

    /**
     * onNameDeleted called for every element/attribute in the document being deleted
     * that matches any of the Indexer pattern.
     *
     * @param pattern Indexer pattern that matches the element/attribute
     * @param key     Document key
     * @param elemID  Element ID that matches the pattern
     * @param attrID  Attribute ID that matches the pattern
     * @throws DBException
     */
    void onNameDeleted(IndexPattern pattern, Key key, short elemID, short attrID) throws DBException;

    /**
     * onValueAdded called for every element/attribute in the document being added
     * that matches any of the Indexer pattern.
     *
     * @param pattern Indexer pattern that matches the element/attribute
     * @param value   Element/attribute value to add
     * @param key     Document key
     * @param pos     The offset into the stream the Element occurs at
     * @param len     The length of the substream for the Element
     * @param elemID  Element ID that matches the pattern
     * @param attrID  Attribute ID that matches the pattern
     * @throws DBException
     */
    void onValueAdded(IndexPattern pattern, String value, Key key, int pos, int len, short elemID, short attrID) throws DBException;

    /**
     * onValueDeleted called for every element/attribute in the document being deleted
     * that matches any of the Indexer pattern.
     *
     * @param pattern Indexer pattern that matches the element/attribute
     * @param value   Element/attribute value to delete
     * @param key     Document key
     * @param pos     The offset into the stream the Element occurs at
     * @param len     The length of the substream for the Element
     * @param elemID  Element ID that matches the pattern
     * @param attrID  Attribute ID that matches the pattern
     * @throws DBException
     */
    void onValueDeleted(IndexPattern pattern, String value, Key key, int pos, int len, short elemID, short attrID) throws DBException;
}
