/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: NullReader.java 541508 2007-05-25 01:54:12Z vgritsenko $
 */

package org.apache.xindice.core.meta.inline;

import org.apache.xindice.core.FaultCodes;
import org.apache.xindice.core.data.Value;

/**
 * Read metadata of length zero.  Handy for comparing performance
 * of the metadata-free database with a database using the metadata
 * machinery but containing no metadata.
 *
 * @version $Revision: 541508 $, $Date: 2007-05-24 21:54:12 -0400 (Thu, 24 May 2007) $
 */
public class NullReader implements InlineMetaReader {

    /**
     * @see org.apache.xindice.core.meta.inline.InlineMetaReader#getVersion()
     */
    public int getVersion() {
        return 0;
    }

    /**
     * @see org.apache.xindice.core.meta.inline.InlineMetaReader#read(Value)
     */
    public InlineMetaMap read(Value data) throws InlineMetaException {
        if (data.getLength() != 0) {
            throw new InlineMetaException(FaultCodes.COL_DOCUMENT_MALFORMED,
                                          "Expecting header length of 0");
        }

        return new NullMap();
    }

    /**
     * A VERY simple and efficient InlineMetaMap implementation
     * that holds precisely nothing.
     */
    public static class NullMap implements InlineMetaMap {

        private static String[] keys = new String[]{};

        /**
         * @see org.apache.xindice.core.meta.inline.InlineMetaMap#containsKey(String)
         */
        public boolean containsKey(String key) {
            return false;
        }

        /**
         * @see org.apache.xindice.core.meta.inline.InlineMetaMap#get(String)
         */
        public Object get(String key) throws InlineMetaException {
            throw new InlineMetaException(FaultCodes.GEN_CRITICAL_ERROR,
                                          "NullMap does not accept key '" + key + "'");
        }

        /**
         * @see org.apache.xindice.core.meta.inline.InlineMetaMap#keys()
         */
        public String[] keys() {
            return keys;
        }

        /**
         * @see org.apache.xindice.core.meta.inline.InlineMetaMap#put(String,Object)
         */
        public void put(String key, Object value) throws InlineMetaException {
            throw new InlineMetaException(FaultCodes.GEN_CRITICAL_ERROR,
                                          "NullMap does not accept key '" + key + "'");
        }
    }
}
