/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: NodeReader.java 584476 2007-10-14 02:41:50Z natalia $
 */

package org.apache.xindice.core.query.ftsearch;

import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.TermFreqVector;
import org.apache.lucene.index.TermEnum;
import org.apache.lucene.index.Term;
import org.apache.lucene.index.TermDocs;
import org.apache.lucene.index.TermPositions;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.FieldSelector;
import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.TokenStream;
import org.apache.lucene.analysis.Token;
import org.apache.lucene.search.DefaultSimilarity;
import org.apache.xindice.xml.dom.NodeImpl;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;

import java.io.IOException;
import java.io.StringReader;
import java.util.Collection;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.HashSet;
import java.util.List;

/**
 * Implementation of IndexReader that works with set of DOM nodes in memory.
 * Set of nodes is constant - no modifications of it are allowed.
 *
 * @version $Revision: 584476 $, $Date: 2007-10-13 22:41:50 -0400 (Sat, 13 Oct 2007) $
 */
public class NodeReader extends IndexReader {
    // list of DOM nodes
    private final ArrayList[] nodes;

    private byte[] norms;

    // maps term to a list of documents where it appears
    private final SortedMap termMap;

    protected NodeReader(NodeList list, Analyzer analyzer) {
        super(null);

        nodes = new ArrayList[list.getLength()];

        for (int i = 0; i < nodes.length; i++) {
            nodes[i] = new ArrayList();
            Node node = list.item(i);

            String text;
            if (node instanceof NodeImpl) {
                // DOM Level 3 method
                text = ((NodeImpl) node).getTextContent();
            } else {
                text = getTextContent(node);
            }

            TokenStream stream = analyzer.tokenStream("", new StringReader(text));
            try {
                Token token;
                while ((token = stream.next()) != null) {
                    nodes[i].add(token.termText());
                }
            } catch (IOException e) {
                // won't happen
            }
        }

        // init norms
        norms = new byte[nodes.length];
        Arrays.fill(norms, DefaultSimilarity.encodeNorm(1.0f));

        // build term enumeration
        termMap = buildTermMap();
    }

    public TermFreqVector[] getTermFreqVectors(int docNumber) {
        throw new UnsupportedOperationException();
    }

    public TermFreqVector getTermFreqVector(int docNumber, String field) {
        throw new UnsupportedOperationException();
    }

    public int numDocs() {
        return nodes.length;
    }

    public int maxDoc() {
        return nodes.length;
    }

    /**
     * Method is not supported.
     */
    public Document document(int n, FieldSelector fieldSelector) {
        return null;
    }

    /**
     * Deletion is not supported.
     */
    public boolean isDeleted(int n) {
        return false;
    }

    /**
     * Deletion is not supported.
     */
    public boolean hasDeletions() {
        return false;
    }

    public byte[] norms(String field) throws IOException {
        return field.length() == 0 ? norms : null;
    }

    public void norms(String field, byte[] bytes, int offset) {
        System.arraycopy(norms, 0, bytes, offset, maxDoc());
    }

    protected void doSetNorm(int doc, String field, byte value) {
        if (field.length() > 0) {
            return;
        }

        norms[doc] = value;
    }

    public TermEnum terms() {
        return new NodeTermEnum(termMap);
    }

    public TermEnum terms(Term t) {
        return new NodeTermEnum(termMap, t);
    }

    /**
     * Builds the map of all the terms in all the nodes to the list of
     * node numbers where those terms appear.
     * @return Map with keys sorted in ascending order
     */
    private SortedMap buildTermMap() {
        SortedMap map = new TreeMap();

        for (int i = 0; i < nodes.length; i++) {
            for (int j = 0; j < nodes[i].size(); j++) {
                String term = (String) nodes[i].get(j);

                List docs;
                if (map.containsKey(term)) {
                    docs = (List) map.get(term);
                } else {
                    docs = new ArrayList();
                }

                docs.add(new Integer(i));
                map.put(term, docs);
            }
        }

        return map;
    }

    public int docFreq(Term t) {
        List docs = (List) termMap.get(t.text());

        // no such term
        if (docs == null) {
            return 0;
        }

        HashSet set = new HashSet();
        set.addAll(docs);
        return set.size();
    }

    public TermDocs termDocs() {
        return new NodeTermDocs(this);
    }

    public TermPositions termPositions() {
        return new NodeTermPositions(this);
    }

    /**
     * Deletion is not supported.
     */
    protected void doDelete(int docNum) {
        throw new UnsupportedOperationException();
    }

    /**
     * Deletion is not supported.
     */
    protected void doUndeleteAll() {
        throw new UnsupportedOperationException();
    }

    /**
     * Not applicable
     */
    protected void doCommit() {
    }

    /**
     * Not applicable
     */
    protected void doClose() {
    }

    /**
     * Field names are not supported.
     */
    public Collection getFieldNames(FieldOption fldOption) {
        throw new UnsupportedOperationException();
    }

    ArrayList[] getNodes() {
        return nodes;
    }

    SortedMap getTermMap() {
        return termMap;
    }

    /**
     * Get text content of a DOM node. This is the same as DOM Level 3 method
     * getTextContent().
     * @param node
     * @return The text content of this node and its descendants.
     */
    private String getTextContent(Node node) {
        String text = null;
        switch (node.getNodeType()) {
            case Node.ATTRIBUTE_NODE:
            case Node.CDATA_SECTION_NODE:
            case Node.COMMENT_NODE:
            case Node.PROCESSING_INSTRUCTION_NODE:
            case Node.TEXT_NODE:
                text = node.getNodeValue();
                break;
            case Node.ELEMENT_NODE:
            case Node.DOCUMENT_FRAGMENT_NODE:
            case Node.ENTITY_NODE:
            case Node.ENTITY_REFERENCE_NODE:
                StringBuffer val = new StringBuffer();

                NodeList children = node.getChildNodes();
                if (children == null || children.getLength() == 0) {
                    text = "";
                    break;
                }

                for (int i = 0; i < children.getLength(); i++) {
                    val.append(getTextContent(children.item(i)));
                }
                text = val.toString();
                break;
        }

        return text;
    }
}
