/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: AddDocument.java 578602 2007-09-23 20:33:31Z natalia $
 */

package org.apache.xindice.tools.command;

import org.apache.xindice.tools.XMLTools;

import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;
import org.xmldb.api.DatabaseManager;
import org.xmldb.api.base.Collection;
import org.xmldb.api.base.Resource;

import javax.xml.parsers.SAXParserFactory;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;

/**
 * AddDocument is designed to let the user add single document
 * and/or multiple documents to any existing collection
 * within the current database.
 *
 * @version $Revision: 578602 $, $Date: 2007-09-23 16:33:31 -0400 (Sun, 23 Sep 2007) $
 */
public class AddDocument extends Command {

    /**
     * Adds a document to the collection
     */
    public boolean execute(XMLTools.Config table) throws Exception {

        if (table.getString(XMLTools.COLLECTION) == null) {
            System.out.println("ERROR : Collection and switch required");
            return false;
        }

        if ("".equals(table.getString(XMLTools.FILE_PATH))) {
            System.out.println("ERROR : File path required");
            return false;
        }

        Collection col = null;
        InputStream fis = null;
        try {
            // Create a collection instance
            String colstring = normalizeCollectionURI(table.getString(XMLTools.COLLECTION),
                                                      table.getBoolean(XMLTools.LOCAL));
            col = DatabaseManager.getCollection(colstring);
            if (col == null) {
                System.out.println("ERROR : Collection not found!");
                return false;
            }

            File file = new File(table.getString(XMLTools.FILE_PATH));
            fis = new FileInputStream(file);

            // Parse in XML using Xerces
            SAXParserFactory spf = javax.xml.parsers.SAXParserFactory.newInstance();
            spf.setNamespaceAware(true);

            XMLReader reader = spf.newSAXParser().getXMLReader();
            StringSerializer ser = new StringSerializer();
            reader.setContentHandler(ser);
            reader.setProperty("http://xml.org/sax/properties/lexical-handler", ser);
            reader.parse(new InputSource(fis));

            // Create the XMLResource and store the document
            Resource resource = col.createResource(table.getString(XMLTools.NAME_OF), "XMLResource");
            resource.setContent(ser.toString());
            col.storeResource(resource);

            System.out.println("Added document " + table.getString(XMLTools.COLLECTION) + "/" + resource.getId());
        } finally {
            if (col != null) {
                col.close();
            }

            if (fis != null) {
                fis.close();
            }
        }

        return true;
    }

    public void usage() {
        System.out.println("Format: xindice ad  -c <context> [-l [-d <path>]] [-v] [parameters...]");
        System.out.println();
        System.out.println("Adds a single document to an existing collection");
        System.out.println();
        System.out.println("Command-specific switches:");
        System.out.println("    -f|--filepath <file>");
        System.out.println("                 File path for the document being added to the collection");
        System.out.println("    -n|--nameOf <name>");
        System.out.println("                 Document key. If this parameter is omitted, database will");
        System.out.println("                 generate an unique key");
        System.out.println();
        System.out.println("Examples:");
        System.out.println("    xindice ad -c /db/test -f /tmp/xmldocument -n myxmldocument");
        System.out.println("    xindice ad -c /db/test/nested -f /tmp/data.xml");
        System.out.println();
    }
}
