/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: StringSerializer.java 541508 2007-05-25 01:54:12Z vgritsenko $
 */

package org.apache.xindice.tools.command;

import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.ext.LexicalHandler;

import java.util.HashMap;
import java.util.Map;
import java.util.Stack;

/**
 * SAX content handler that produces a string from the SAX events it receives.
 * After calling <code>endDocument()</code>, the string becomes available by
 * calling <code>toString()</code>
 *
 * @author <a href="mailto:james.bates@amplexor.com">James Bates</a>
 * @version $Revision: 541508 $, $Date: 2007-05-24 21:54:12 -0400 (Thu, 24 May 2007) $
 */
public class StringSerializer implements ContentHandler, LexicalHandler {

    /**
     * The encoding that should be written into the XML declaration. May be
     * <code>null</code> indicating no encoding should be written.
     */
    private String encoding;

    /**
     * Map containing stack of prefixes used in the XML document
     */
    private Map namespaceMap;

    /**
     * Namespace declarations that should be written out when the
     * next element opens
     */
    private StringBuffer namespaceDecls;

    /**
     * Are we currently inside a CDATA section?
     */
    private boolean inCDATA;

    /**
     * The document so far
     */
    private StringBuffer outputXml;

    /**
     * Creates new <code>StringSerializer</code>. The output encoding is not
     * specified therefore will be treated as "UTF-8".
     */
    public StringSerializer() {
        this(null);
    }

    /**
     * Creates new <code>StringSerializer</code>.
     *
     * @param encoding The encoding that should be written into the
     *                 XML declaration for the document. This encoding
     *                 is not used in any other way: when the resultant
     *                 string is written to a file, the writing application
     *                 must still take care to actually write using the
     *                 correct encoding. If <code>encoding</code> is
     *                 <code>null</code>, the encoding will be omitted
     *                 from the XML declaration.
     */
    public StringSerializer(String encoding) {
        this.encoding = encoding;
        this.outputXml = new StringBuffer(1024); // allocate a 1k block to start with
        this.namespaceMap = new HashMap();
        this.namespaceDecls = new StringBuffer();
    }

    public void startDocument() throws SAXException {
        outputXml.append("<?xml version=\"1.0\"");
        if (encoding != null) {
            outputXml.append(" encoding=\"");
            outputXml.append(encoding);
            outputXml.append('\"');
        }
        outputXml.append("?>");
    }

    public void endDocument() throws SAXException {
    }

    public void processingInstruction(String target, String data) throws SAXException {
        outputXml.append("<?");
        outputXml.append(target);
        outputXml.append(' ');
        outputXml.append(data);
        outputXml.append("?>");
    }

    public void comment(char[] data, int start, int len) throws SAXException {
        outputXml.append("<!--");
        outputXml.append(data, start, len);
        outputXml.append("-->");
    }

    private void xmlIze(StringBuffer sb,
                        char[] data, int start, int len,
                        boolean isAttValue) {

        for (int i = start; i < start + len; i++) {
            if (data[i] == '<') {
                sb.append("&lt;");
            } else if (data[i] == '&') {
                sb.append("&amp;");
            } else if ((data[i] == '"') && (isAttValue)) {
                sb.append("&quot;");
            } else {
                sb.append(data[i]);
            }
        }
    }

    private void writeAttribute(StringBuffer sb, String attName, String attValue) {
        sb.append(' ');
        sb.append(attName);
        sb.append("=\"");
        xmlIze(sb, attValue.toCharArray(), 0, attValue.length(), true);
        sb.append('\"');
    }

    public void startPrefixMapping(String prefix, String uri) throws SAXException {
        Stack uriStack = (Stack) namespaceMap.get(prefix);
        if (uriStack == null) {
            uriStack = new Stack();
            namespaceMap.put(prefix, uriStack);
        }

        uriStack.push(uri);

        if (!prefix.equals("")) {
            writeAttribute(namespaceDecls, "xmlns:" + prefix, uri);
        } else {
            writeAttribute(namespaceDecls, "xmlns", uri);
        }
    }

    public void endPrefixMapping(String prefix) throws SAXException {
        Stack uriStack = (Stack) namespaceMap.get(prefix);
        uriStack.pop();
    }

    public void ignorableWhitespace(char[] data, int start, int len) throws SAXException {
        outputXml.append(data, start, len);
    }

    private String getPrefix(String qName) {
        if (qName.indexOf(':') != -1) {
            return qName.substring(0, qName.indexOf(':'));
        } else {
            return "";
        }
    }

    public void startElement(String ns, String localName, String qName, Attributes att)
            throws SAXException {

        /* First: check element name qualification */
        String prefix = getPrefix(qName);

        if ((!(prefix.equals("") || prefix.equals("xml")))
                && namespaceMap.get(prefix) == null) {

            writeAttribute(namespaceDecls, "xmlns:" + prefix, ns);
        }

        outputXml.append('<');
        outputXml.append(qName);
        outputXml.append(namespaceDecls);
        namespaceDecls = new StringBuffer();

        for (int i = 0; i < att.getLength(); i++) {

            String attNamespaceUri = att.getURI(i);
            String attQName = att.getQName(i);
            String attValue = att.getValue(i);

            String attPrefix = getPrefix(attQName);
            if (!(attPrefix.equals("") || attPrefix.equals("xml"))) {

                if (namespaceMap.get(attPrefix) == null) {
                    writeAttribute(outputXml, "xmlns:" + attPrefix, attNamespaceUri);
                }
            }

            writeAttribute(outputXml, attQName, attValue);
        }

        outputXml.append('>');
    }

    public void endElement(String ns, String localName, String qName)
            throws SAXException {
        outputXml.append("</");
        outputXml.append(qName);
        outputXml.append('>');
    }

    public void skippedEntity(String entity) throws SAXException {
        if (entity.startsWith("%")) {
            outputXml.append(entity);
            outputXml.append(';');
        } else {
            outputXml.append('&');
            outputXml.append(entity);
            outputXml.append(';');
        }
    }

    public void characters(char[] data, int start, int len) throws SAXException {
        if (inCDATA) {
            outputXml.append(data, start, len);
        } else {
            xmlIze(outputXml, data, start, len, false);
        }
    }

    public void setDocumentLocator(org.xml.sax.Locator locator) {
    }

    public void startCDATA() throws SAXException {
        outputXml.append("<![CDATA[");
        inCDATA = true;
    }

    public void endCDATA() throws SAXException {
        inCDATA = false;
        outputXml.append("]]>");
    }

    public void startEntity(String entity) throws SAXException {
    }

    public void endEntity(String entity) throws SAXException {
    }

    public void startDTD(String docType, String systemID, String publicID) throws SAXException {
    }

    public void endDTD() throws SAXException {
    }

    public String toString() {
        return outputXml.toString();
    }
}
