/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: PartialResponse.java 551439 2007-06-28 04:10:20Z vgritsenko $
 */

package org.apache.xindice.webadmin;

import org.apache.xindice.util.StringUtilities;

import java.util.ArrayList;

/**
 * PartialResponse represents response part of the multi-status response
 * document. Every partial response has its correspondent URL information
 * (href element) and one or more elements with additional request-specific
 * information.
 *
 * @version $Revision: 551439 $, $Date: 2007-06-28 00:10:20 -0400 (Thu, 28 Jun 2007) $
 */
public class PartialResponse {
    private final String href;
    private ArrayList children;

    public PartialResponse(String href) {
        this.href = href;
    }

    public String getHref() {
        return href;
    }

    /**
     * Adds another element to the list of children of this partial response.
     *
     * @param name Element name
     * @param value Element value
     */
    public void addContent(String name, String value) {
        Content child = new Content(name);
        child.setValue(value);
        addContent(child);
    }

    /**
     * Adds another element to the list of children of this partial response.
     *
     * @param child Existing element
     */
    public void addContent(Content child) {
        if (children == null) {
            children = new ArrayList();
        }
        children.add(child);
    }

    /**
     * Builds XML out of partial response object.
     *
     * @return XML string
     */
    public String toString() {
        StringBuffer buf = new StringBuffer(1024);
        buf.append("<response>");
        buf.append("<href>").append(StringUtilities.escapeXml(href)).append("</href>");

        if (children != null && children.size() != 0) {
            for (int i = 0; i < children.size(); i++) {
                buf.append(((Content) children.get(i)).getContent());
            }
        }

        buf.append("</response>");

        return buf.toString();
    }

    /**
     * Content class holds request-specific information in a tree-like
     * structure. Every element of this tree can either be empty, have a
     * text value or hold children elements.
     */
    public static class Content {
        private ArrayList children;
        private String name;
        private String value;

        public Content(String name) {
            this.name = name;
        }

        public Content(String name, String value) {
            this.name = name;
            this.value = value;
        }

        public void setValue(String value) {
            this.value = value;
        }

        public Content addChild(String name) {
            Content child = new Content(name);
            addChild(child);
            return child;
        }

        public Content addChild(String name, String value) {
            Content child = addChild(name);
            child.setValue(value);
            return child;
        }

        public void addChild(Content child) {
            if (children == null) {
                children = new ArrayList();
            }
            children.add(child);
        }

        public String getValue() {
            return value;
        }

        private StringBuffer getContent() {
            StringBuffer buf = new StringBuffer(1024);

            buf.append("<").append(name).append(">");
            if (value != null) {
                buf.append(StringUtilities.escapeXml(value));
            } else if (children != null && children.size() != 0) {
                for (int i = 0; i < children.size(); i++) {
                    buf.append(((Content) children.get(i)).getContent());
                }
            }

            buf.append("</").append(name).append(">");

            return buf;
        }

        public String toString() {
            return getContent().toString();
        }
    }
}
