/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: MimeTable.java 541515 2007-05-25 02:45:06Z vgritsenko $
 */

package org.apache.xindice.webadmin.util;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.Writer;
import java.util.Hashtable;
import java.util.HashMap;
import java.util.Iterator;

import org.apache.xindice.xml.dom.DOMParser;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;

/**
 * MimeTable maps a file extension to a mime type.
 *
 * @author <a href="mailto:jmetzner@apache.org">Jan Metzner</a>
 * @version $Revision: 541515 $, $Date: 2007-05-24 22:45:06 -0400 (Thu, 24 May 2007) $
 */
public class MimeTable {

    /** the default mime type for binaries */
    public static final String BINARY_MIME_TYPE = "application/octet-stream";

    /** the default mime type for xml documents */
    public static final String XML_MIME_TYPE = "text/xml";

    /** holds the extensions/mime types */
    protected static Hashtable mimeMappings = new Hashtable();

    private final static Log log = LogFactory.getLog(MimeTable.class);

    /**
     * Configures the Mime table.
     * the configuration can look like the Configuration in the
     * addMimeConfig(Configuration config) Method<br />
     * <br/>
     * or if the Configuration should be loaded out of an extra file:
     * <pre>
     * &lt;mime location="/some/path"&gt;
     * </pre>
     * or if the Configuration should be loaded out of the Classpath:
     * <pre>
     * &lt;mime location="resource://org.a.MyFile"&gt;
     * </pre>
     */
    public static void configure(String location) {
        // load mime table
        boolean loadFromClasspath = false;
        if (location.startsWith("resource://")) {
            loadFromClasspath = true;
            location = location.substring(11);
        }

        try {
            if (loadFromClasspath) {
                InputStream is = Thread.currentThread().getContextClassLoader().getResource(location).openStream();
                if (log.isDebugEnabled()) {
                    log.debug("Load Configuration from Classpath: " + location);
                }
                addMimeConfig(DOMParser.toDocument(is));
            } else {
                if (log.isDebugEnabled()) {
                    log.debug("Load Configuration from File: " + location);
                }
                addMimeConfig(DOMParser.toDocument(location));
            }
        } catch (Exception e) {
            log.error("Could not load MimeTable configuration", e);
        }
    }

    /**
     * Add the given mime Document to this mime table.
     * the mime Document looks like this:
     * <pre>
     * &lt;mime&gt;
     *  &lt;mime-mapping&gt;
     *   &lt;extension&gt;xml&lt;/extension&gt;
     *   &lt;mime-type&gt;xml&lt;/mime-type&gt;
     *  &lt;/mime-mapping&gt;
     * ...
     * &lt;/mime&gt;
     * </pre>
     *
     * @param doc mime config.
     */
    public static void addMimeConfig(Document doc) {
        NodeList mapping = doc.getChildNodes();
        for (int i = 0; i < mapping.getLength(); i++) {
            NodeList thisMapping = mapping.item(i).getChildNodes();
            String ext = "";
            String type = "";
            for (int j = 0; j < thisMapping.getLength(); j++) {
                if (thisMapping.item(j).getNodeName().equalsIgnoreCase("extension")) {
                    ext = thisMapping.item(j).getNodeValue();
                } else if (thisMapping.item(j).getNodeName().equalsIgnoreCase("mime-type")) {
                    type = thisMapping.item(j).getNodeValue();
                }
            }
            mimeMappings.put(ext, type);
        }
    }

    /**
     * Returns the mime type for the given name.
     * If no mime for this name is available the parameter defaultXml specifies if
     * the xml mime type or the binary mime type are returned.
     *
     * @param resourceName
     * @return mime type for this resource name
     */
    public static String getMimeType(String resourceName) {
        int split = resourceName.lastIndexOf('.');
        String mime = null;
        if (split > 0) {
            mime = (String) mimeMappings.get(resourceName.substring(split + 1));
        }

        if (mime != null) {
            return mime;
        } else {
            return BINARY_MIME_TYPE;
        }
    }

    private static void writeMappingAsXml(HashMap mappings, Writer out) throws IOException {
        out.write("<?xml version=\"1.0\"?>\n");
        out.write("<mime>\n");
        for (Iterator i = mappings.keySet().iterator(); i.hasNext();) {
            String ext = (String) i.next();
            String type = (String) mappings.get(ext);

            out.write("    <mime-mapping>\n");
            out.write("        <extension>" + ext + "</extension>\n");
            out.write("        <mime-type>" + type + "</mime-type>\n");
            out.write("    </mime-mapping>\n");
        }
        out.write("</mime>");
    }

    private static HashMap parseApacheHttpdMime(String location) throws IOException {
        HashMap mappings = new HashMap();
        BufferedReader f = new BufferedReader(new FileReader(location));
        String line;
        while ((line = f.readLine()) != null) {
            int startComment = line.indexOf('#');
            if (startComment >= 0) {
                line = line.substring(0, startComment);
            }
            line = line.trim();
            String[] parsedLine = line.split(" |\\t");
            if (parsedLine.length >= 2) {
                for (int i = 1; i < parsedLine.length; i++) {
                    if (parsedLine[i].length() > 0) {
                        mappings.put(parsedLine[i], parsedLine[0]);
                    }
                }
            }
        }
        return mappings;
    }

    private static void convertApacheHttpdMimeToXml(String source, String destination) throws IOException {
        HashMap mappings = parseApacheHttpdMime(source);
        BufferedWriter output = new BufferedWriter(new FileWriter(destination));
        writeMappingAsXml(mappings, output);
    }

    public static void main(String[] args) throws Exception {
        if (args.length != 2) {
            System.out.println("Convert Apache httpd mime file to a xml file");
            System.out.println("Usage: java org.apache.xindice.webadmin.util.MimeTable httpdfile xmlfile");
            System.exit(1);
        }

        convertApacheHttpdMimeToXml(args[0], args[1]);
        System.out.println("Converted " + args[0] + " to " + args[1]);
    }
}
