/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: ListCollectionViewer.java 541515 2007-05-25 02:45:06Z vgritsenko $
 */

package org.apache.xindice.webadmin.viewer.components;

import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.TimeZone;

import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.xindice.core.Collection;
import org.apache.xindice.core.DBException;
import org.apache.xindice.webadmin.viewer.HtmlCollectionViewer;
import org.apache.xindice.webadmin.viewer.HtmlViewerComponent;

/**
 * Xindice Html Viewer for collection listings.
 *
 * @author <a href="mailto:jmetzner@apache.org">Jan Metzner</a>
 * @version $Revision: 541515 $, $Date: 2007-05-24 22:45:06 -0400 (Thu, 24 May 2007) $
 */
public class ListCollectionViewer extends HtmlViewerComponent implements HtmlCollectionViewer {

    /**
     * Simple date format for the modification date representation.
     */
    public static final String DATE_FORMAT = "EEE, dd MMM yyyy HH:mm:ss zzz";

    public void execute(HttpServletRequest req, HttpServletResponse res, Collection col)
            throws ServletException, IOException {

        String path = col.getCanonicalName();
        String title = "List contents for " + path;

        TableRow row = new TableRow();
        ServletOutputStream output = startViewer(title, path, res);
        printStartTable(row.getTitles(), output);

        // write parent
        printTableRow(row.getParentCollection(), output);

        // write child collections
        String[] childCol;
        try {
            childCol = col.listCollections();
        } catch (DBException e) {
            childCol = new String[0];
        }
        for (int i = 0; i < childCol.length; i++) {
            printTableRow(row.getCollection(childCol[i]), output);
        }

        // write child resources
        String[] childRes;
        try { // get names
            childRes = col.listDocuments();
        } catch (Exception e) {
            childRes = new String[0];
        }

        for (int i = 0; i < childRes.length; i++) {
            String date;
            try { // get date
                long modificationDate = col.getEntryMeta(childRes[i]).getModificationTime();
                if (modificationDate != 0) {
                    SimpleDateFormat formatter = new SimpleDateFormat(DATE_FORMAT, Locale.US);
                    formatter.setTimeZone(TimeZone.getTimeZone("GMT"));
                    date = formatter.format(new Date(modificationDate));
                } else {
                    date = "";
                }
            } catch (DBException e) {
                date = "";
            }
            printTableRow(row.getResource(childRes[i], date), output);
        }
        printEndTable(output);
        finishViewer(output);
    }

    protected String getName() {
        return "list";
    }

    // TableRow Inner Class
    protected class TableRow {

        protected final String[] titles = {"Name", "Type", "Last Modification", "Source", "Delete"};

        protected String[] content;

        TableRow() {
            content = new String[titles.length];
            reset();
        }

        protected String[] getTitles() {
            return titles;
        }

        protected void reset() {
            for (int i = 0; i < content.length; i++) {
                content[i] = null;
            }
        }

        protected String[] getCollection(String name) {
            reset();
            content[0] = createLink(name + "/?viewer=list", name);
            content[1] = "collection";
            content[3] = "&nbsp;";
            content[4] = createLink(name + "/?viewer=delete&redirect=../?viewer=list", "delete");
            return content;
        }

        protected String[] getResource(String name, String date) {
            reset();
            content[0] = createLink(name + "?viewer=default", name);
            content[1] = "resource";
            content[2] = date;
            content[3] = createLink(name, "src");
            content[4] = createLink(name + "?viewer=delete&redirect=./?viewer=list", "delete");
            return content;
        }

        protected String[] getParentCollection() {
            reset();
            content[0] = createLink("../?viewer=list", "up to parent collection");
            content[1] = "collection";
            return content;
        }

        protected String createLink(String link, String name) {
            return "<a href=\"" + link + "\">" + name + "</a>";
        }
    }
}
