/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: Mkcol.java 541515 2007-05-25 02:45:06Z vgritsenko $
 */

package org.apache.xindice.webadmin.webdav.components;

import java.io.IOException;

import javax.servlet.ServletException;

import org.apache.xindice.core.Collection;
import org.apache.xindice.core.DBException;
import org.apache.xindice.core.FaultCodes;
import org.apache.xindice.webadmin.util.CollectionConfigurationHelper;
import org.apache.xindice.webadmin.webdav.WebdavStatus;
import org.apache.xindice.webadmin.webdav.DAVResponse;
import org.apache.xindice.webadmin.webdav.DAVRequest;
import org.apache.xindice.webadmin.Location;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * This class implements the Mkcol command for WebDAV operations on
 * Xindice.<br>
 * <br>
 * MKCOL command creates new collection at a specified location. There must be
 * no existing collection at the specified location. Newly created collection
 * will be empty. WebDAV specification allows request to have a body that may
 * contain instructions for creating child resources, but this behavior is
 * not suppoted. <br>
 * <br>
 * Operation possible status codes include: <ul>
 * <li>
 * 201 (Created) - The collection or structured resource was created in
 * its entirety.
 * <li>
 * 403 (Forbidden) - This indicates at least one of two conditions: 1)
 * the server does not allow the creation of collections at the given
 * location in its namespace, or 2) the parent collection of the
 * Request-URI exists but cannot accept members.
 * <li>
 * 405 (Method Not Allowed) - MKCOL can only be executed on a
 * deleted/non-existent resource.
 * <li>
 * 409 (Conflict) - A collection cannot be made at the Request-URI until
 * one or more intermediate collections have been created.
 * <li>
 * 415 (Unsupported Media Type)- The server does not support the request
 * type of the body.
 * </ul>
 *
 * @author <a href="mailto:jmetzner@apache.org">Jan Metzner</a>
 * @author <a href="mailto:gianugo@apache.org">Gianugo Rabellino</a>
 * @version $Revision: 541515 $, $Date: 2007-05-24 22:45:06 -0400 (Thu, 24 May 2007) $
 */
public class Mkcol implements DAVComponent {
    private final Log log = LogFactory.getLog(Mkcol.class);

    public void execute(DAVRequest req, DAVResponse res, Location target) throws ServletException, IOException {
        if (target.isRoot()) {
            res.setStatus(WebdavStatus.SC_METHOD_NOT_ALLOWED);
            return;
        }

        Collection col = target.getCollection();
        String name = target.getName();

        if (name != null && req.getRequestDoc() != null) {
            // request body is not supported
            res.setStatus(WebdavStatus.SC_UNSUPPORTED_MEDIA_TYPE);
            return;
        }

        if (col == null) { // parent collection does not exist
            res.setStatus(WebdavStatus.SC_CONFLICT);
            return;
        }

        if (name == null) {
            if (log.isDebugEnabled()) {
                log.debug("Collection already exists");
            }
            res.setStatus(WebdavStatus.SC_METHOD_NOT_ALLOWED);
        } else {
            try {
                res.setStatus(createCollection(col, name));
            } catch (DBException e) {
                log.error("Failed to create collection " + col.getCanonicalName() + "/" + name, e);
                throw new ServletException(e);
            }
        }
    }

    private int createCollection(Collection parent, String name) throws DBException {
        try {
            parent.createCollection(name, CollectionConfigurationHelper.createDefaultConfiguration(name));
            if (log.isDebugEnabled()) {
                log.debug("Collection " + name + " created");
            }

            // TODO: set SC_FORBIDDEN for system collections
        } catch (DBException e) {
            if (e.faultCode == FaultCodes.COL_DUPLICATE_COLLECTION) {
                return WebdavStatus.SC_METHOD_NOT_ALLOWED;
            } else if (e.faultCode == FaultCodes.COL_COLLECTION_NOT_FOUND) {
                return WebdavStatus.SC_METHOD_NOT_ALLOWED;
            } else if (e.faultCode == FaultCodes.COL_CANNOT_CREATE) {
                return WebdavStatus.SC_FORBIDDEN;
            }

            throw e;
        }

        return WebdavStatus.SC_CREATED;
    }
}
