/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: CompressedNode.java 541508 2007-05-25 01:54:12Z vgritsenko $
 */

package org.apache.xindice.xml.dom;

import org.w3c.dom.DOMException;
import org.w3c.dom.Node;

/**
 * CompressedNode represents a Xindice Compressed Node.  The compression
 * system operates independently of the database and only requires a
 * SymbolTable instance to perform compression/decompression.
 *
 * @version $Revision: 541508 $, $Date: 2007-05-24 21:54:12 -0400 (Thu, 24 May 2007) $
 */
public interface CompressedNode extends Node {

    /**
     * isDirty returns whether or not the current Node (or any of its
     * children) have been modified since being loaded.
     *
     * @return Whether or not the Node has been modified
     */
    boolean isDirty();

    /**
     * getDataBytes returns the byte array that defines this Node.  A
     * Node definition is usually a subset of this byte array.
     *
     * @return The Node definition's byte array
     */
    byte[] getDataBytes();

    /**
     * getDataPos returns the offset into the Node definition's byte
     * array that the Node starts at.
     *
     * @return The Node's position
     */
    int getDataPos();

    /**
     * getDataLen returns the length of the Node's definition in the
     * byte array.
     *
     * @return The Node's length
     */
    int getDataLen();

    /**
     * setDataBytes sets the definition and positional information for
     * the Node.
     *
     * @param data The Node definition's byte array
     * @param pos The Node's position
     * @param len The Node's length
     */
    void setDataBytes(byte[] data, int pos, int len);

    /**
     * setDataBytes sets the byte array that defines this Node.
     *
     * @param data The Node definition's byte array
     */
    void setDataBytes(byte[] data);

    /**
     * setDataPos sets the offset into the Node definition's byte array
     * that the Node starts at.
     *
     * @param pos The Node's position
     */
    void setDataPos(int pos);

    /**
     * setDataLen sets the length of the Node's definition in the
     * byte array.
     *
     * @param len The Node's length
     */
    void setDataLen(int len);

    /**
     * load forces the Node to be graphed for this level.  Loading happens
     * automatically whenever a Node's children are accessed, so load
     * should never have to explicitly be called.
     *
     * @throws DOMException if an exception occurs
     */
    void load() throws DOMException;

    /**
     * unload forces the Node to be unloaded for this level.  Whether the
     * node is actually unloaded or not is up to the underlying
     * implementation, and should silently fail if the implementation
     * doesn't support unloading.
     *
     * @throws DOMException if an exception occurs
     */
    void unload() throws DOMException;

    /**
     * isLoaded returns whether or not the Node's definition has been
     * loaded.
     *
     * @return Whether or not the Node is loaded
     */
    boolean isLoaded();

    /**
     * getSymbolID returns the Symbol ID for the current node.  Some nodes
     * will not have a Symbol ID, in which case, a value of -1 is returned.
     *
     * @return The Symbol ID
     */
    short getSymbolID();
}
