/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: DOMImplementationImpl.java 541508 2007-05-25 01:54:12Z vgritsenko $
 */

package org.apache.xindice.xml.dom;

import org.w3c.dom.DOMException;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;

/**
 * DOMImplementationImpl
 *
 * @version $Revision: 541508 $, $Date: 2007-05-24 21:54:12 -0400 (Thu, 24 May 2007) $
 */
public final class DOMImplementationImpl implements DOMImplementation {

    private static final String[][] FEATURES = {
        {"XML",          "1.0"},
        {"XML",          "2.0"},
        {"Traversal",    "2.0"},
        {"Xindice-DB",   "1.0"}, // DBNode and DBDocument
        {"Xindice-Comp", "1.0"}  // CompressedNode and CompressedDocument
    };

    private static final DOMImplementation IMPL = new DOMImplementationImpl();


    private DOMImplementationImpl() {
    }

    public static DOMImplementation getInstance() {
        return IMPL;
    }

    static boolean HasFeature(String feature, String version) {
        for (int i = 0; i < FEATURES.length; i++) {
            if (FEATURES[i][0].equals(feature) && FEATURES[i][1].equals(version)) {
                return true;
            }
        }

        return false;
    }

    /**
     * Test if the DOM implementation implements a specific feature.
     * @param feature The package name of the feature to test. In Level 1, the
     *   legal values are "HTML" and "XML" (case-insensitive).
     * @param version This is the version number of the package name to test.
     *   In Level 1, this is the string "1.0". If the version is not specified,
     *   supporting any version of the  feature will cause the method to return
     *   <code>true</code>.
     * @return <code>true</code> if the feature is implemented in the specified
     *   version, <code>false</code> otherwise.
     */
    public boolean hasFeature(String feature, String version) {
        return HasFeature(feature, version);
    }

    /**
     * Creates an empty <code>DocumentType</code> node. Entity declarations
     * and notations are not made available. Entity reference expansions and
     * default attribute additions do not occur. It is expected that a
     * future version of the DOM will provide a way for populating a
     * <code>DocumentType</code>.
     * <br>HTML-only DOM implementations do not need to implement this method.
     * @param qualifiedName The qualified name of the document type to be
     *   created.
     * @param publicId The external subset public identifier.
     * @param systemId The external subset system identifier.
     * @return A new <code>DocumentType</code> node with
     *   <code>Node.ownerDocument</code> set to <code>null</code>.
     * @exception DOMException
     *   INVALID_CHARACTER_ERR: Raised if the specified qualified name
     *   contains an illegal character.
     *   <br>NAMESPACE_ERR: Raised if the <code>qualifiedName</code> is
     *   malformed.
     * @since DOM Level 2
     */
    public DocumentType createDocumentType(String qualifiedName, String publicId, String systemId) throws DOMException {
        return new DocumentTypeImpl();
    }

    /**
     * Creates an XML <code>Document</code> object of the specified type with
     * its document element. HTML-only DOM implementations do not need to
     * implement this method.
     * @param namespaceURI The namespace URI of the document element to create.
     * @param qualifiedName The qualified name of the document element to be
     *   created.
     * @param doctype The type of document to be created or <code>null</code>.
     *   When <code>doctype</code> is not <code>null</code>, its
     *   <code>Node.ownerDocument</code> attribute is set to the document
     *   being created.
     * @return A new <code>Document</code> object.
     * @exception DOMException
     *   INVALID_CHARACTER_ERR: Raised if the specified qualified name
     *   contains an illegal character.
     *   <br>NAMESPACE_ERR: Raised if the <code>qualifiedName</code> is
     *   malformed, if the <code>qualifiedName</code> has a prefix and the
     *   <code>namespaceURI</code> is <code>null</code>, or if the
     *   <code>qualifiedName</code> has a prefix that is "xml" and the
     *   <code>namespaceURI</code> is different from "
     *   http://www.w3.org/XML/1998/namespace" .
     *   <br>WRONG_DOCUMENT_ERR: Raised if <code>doctype</code> has already
     *   been used with a different document or was created from a different
     *   implementation.
     * @since DOM Level 2
     */
    public Document createDocument(String namespaceURI, String qualifiedName, DocumentType doctype) throws DOMException {
        DocumentImpl doc = new DocumentImpl();
        Element elem = doc.createElement(qualifiedName);
        doc.appendChild(elem);
        if (namespaceURI != null) {
            if (qualifiedName.indexOf(":") != -1) {
                elem.setAttribute("xmlns:" + qualifiedName.substring(0, qualifiedName.indexOf(":")), namespaceURI);
            } else {
                elem.setAttribute("xmlns", namespaceURI);
            }
        }
        if (doctype != null) {
            if (doctype.getOwnerDocument() != null
                    || !(doctype instanceof DocumentTypeImpl)) {
                throw NodeImpl.EX_WRONG_DOCUMENT;
            }
            doc.setDoctype(doctype);
        }
        return doc;
    }

    /**
     * @since DOM Level 3
     */
    public Object getFeature(String feature, String version) {
        if (hasFeature(feature, version)) {
            return this;
        }

        return null;
    }
}

