/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: DocumentTypeImpl.java 541508 2007-05-25 01:54:12Z vgritsenko $
 */

package org.apache.xindice.xml.dom;

import org.apache.xindice.util.StringUtilities;

import org.w3c.dom.DocumentType;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

/**
 * DocumentTypeImpl
 *
 * @version $Revision: 541508 $, $Date: 2007-05-24 21:54:12 -0400 (Thu, 24 May 2007) $
 */
public final class DocumentTypeImpl extends NodeImpl
                                    implements DocumentType {

    public DocumentTypeImpl() {
        super(null, true);
    }

    public short getNodeType() {
        return Node.DOCUMENT_TYPE_NODE;
    }

    /**
     * The name of DTD; i.e., the name immediately following the
     * <code>DOCTYPE</code> keyword.
     */
    public String getName() {
        return nodeName;
    }

    /**
     * A <code>NamedNodeMap</code> containing the general entities, both
     * external and internal, declared in the DTD. Duplicates are discarded.
     * For example in:&lt;!DOCTYPE ex SYSTEM "ex.dtd" [ &lt;!ENTITY foo
     * "foo"&gt; &lt;!ENTITY bar "bar"&gt; &lt;!ENTITY % baz "baz"&gt;]&gt;
     * &lt;ex/&gt;  the interface provides access to <code>foo</code> and
     * <code>bar</code> but not <code>baz</code>. Every node in this map also
     * implements the <code>Entity</code> interface.
     * <br>The DOM Level 1 does not support editing entities, therefore
     * <code>entities</code> cannot be altered in any way.
     */
    public NamedNodeMap getEntities() {
        return null;
    }

    /**
     * A <code>NamedNodeMap</code> containing  the notations declared in the
     * DTD. Duplicates are discarded. Every node in this map also implements
     * the <code>Notation</code> interface.
     * <br>The DOM Level 1 does not support editing notations, therefore
     * <code>notations</code> cannot be altered in any way.
     */
    public NamedNodeMap getNotations() {
        return null;
    }

    /**
     * The public identifier of the external subset.
     * @since DOM Level 2
     */
    public String getPublicId() {
        return null;
    }

    /**
     * The system identifier of the external subset.
     * @since DOM Level 2
     */
    public String getSystemId() {
        return null;
    }

    /**
     * The internal subset as a string.The actual content returned depends on
     * how much information is available to the implementation. This may
     * vary depending on various parameters, including the XML processor
     * used to build the document.
     * @since DOM Level 2
     */
    public String getInternalSubset() {
        return null;
    }

    //
    // DOM Level 3 Implementation
    //

    /**
     * @since DOM Level 3
     */
    public String getTextContent() {
        return null;
    }

    /**
     * Tests whether two nodes are equal.
     * This method tests for equality of nodes, not sameness (i.e., whether
     * the two nodes are references to the same object) which can be tested with
     * <code>Node.isSameNode()</code>. All nodes that are the same will also be
     * equal, though the reverse may not be true.
     * <br>
     * Two nodes are equal if and only if the following conditions are satisfied:
     * <ul><li>The two nodes are of the same type.</li>
     * <li>The following string attributes are equal: nodeName, localName,
     * namespaceURI, prefix, nodeValue. This is: they are both null, or they have
     * the same length and are character for character identical.</li>
     * <li>The attributes <code>NamedNodeMaps</code> are equal. This is: they are both null,
     * or they have the same length and for each node that exists in one map there
     * is a node that exists in the other map and is equal, although not
     * necessarily at the same index.</li>
     * <li>The childNodes NodeLists are equal. This is: they are both null, or they
     * have the same length and contain equal nodes at the same index. Note that
     * normalization can affect equality; to avoid this, nodes should be normalized
     * before being compared.</li></ul>
     * <br>
     * For two DocumentType nodes to be equal, the following conditions must also
     * be satisfied:
     * <ul><li>The following string attributes are equal: publicId, systemId,
     * internalSubset.</li>
     * <li>The entities <code>NamedNodeMaps</code> are equal.</li>
     * <li>The notations <code>NamedNodeMaps</code> are equal.</li></ul>
     * <br>
     * On the other hand, the following do not affect equality: the ownerDocument,
     * baseURI, and parentNode attributes, the specified attribute for Attr nodes,
     * the schemaTypeInfo attribute for Attr and Element nodes,
     * the Text.isElementContentWhitespace attribute for Text nodes, as well as any
     * user data or event listeners registered on the nodes.
     * @param other Node to test againts
     * @return true if nodes are equal, false otherwise
     * @since DOM Level 3
     */
    public boolean isEqualNode(Node other) {
        if (!super.isEqualNode(other)) {
            return false;
        }

        DocumentType otherDocType = (DocumentType) other;
        if (!(StringUtilities.equals(this.getPublicId(), otherDocType.getPublicId()) &&
              StringUtilities.equals(this.getSystemId(), otherDocType.getSystemId()) &&
              StringUtilities.equals(this.getInternalSubset(), otherDocType.getInternalSubset()))) {
            return false;
        }

        NamedNodeMap ent = getEntities();
        NamedNodeMap otherEnt = otherDocType.getEntities();
        if (!equalNodeMaps(ent, otherEnt)) {
            return false;
        }

        NamedNodeMap not = getNotations();
        NamedNodeMap otherNot = otherDocType.getNotations();

        return equalNodeMaps(not, otherNot);
    }

    private boolean equalNodeMaps(NamedNodeMap map1, NamedNodeMap map2) {
        if (map1 == null && map2 == null) {
            return true;
        }
        if (map1 == null || map2 == null) {
            return false;
        }

        if (map1.getLength() != map2.getLength()) {
            return false;
        }

        for (int i = 0; i < map1.getLength(); i++) {
            if (!((NodeImpl) map1.item(i)).isEqualNode(map2.item(i))) {
                return false;
            }
        }

        return true;
    }
}
