/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: DatabaseTest.java 530223 2007-04-19 01:48:06Z vgritsenko $
 */

package org.apache.xindice.core;

import org.apache.xindice.util.Configuration;
import org.apache.xindice.xml.dom.DOMParser;

import junit.framework.TestCase;

/**
 * Tests Xindice Core API (org.apache.xindice.core.Database,
 * org.apache.xindice.core.Collection) createCollection.
 *
 * @author <a href="mailto:vgritsenko@apache.org">Vadim Gritsenko</a>
 * @version $Revision: 530223 $, $Date: 2007-04-18 21:48:06 -0400 (Wed, 18 Apr 2007) $
 */
public class DatabaseTest extends TestCase {

    public static final String DATABASE =
        "<root-collection dbroot='db/' name='db' use-metadata='on'>" +
            "<queryengine>" +
                "<resolver autoindex='false' class='org.apache.xindice.core.query.XPathQueryResolver'/>" +
                "<resolver class='org.apache.xindice.core.xupdate.XUpdateQueryResolver'/>" +
            "</queryengine>" +
        "</root-collection>";

    public static final String COLLECTIONA =
        "<collection compressed='true' name='CollectionA'>" +
            "<filer class='org.apache.xindice.core.filer.BTreeFiler'/>" +
            "<indexes/>" +
        "</collection>";

    public static final String COLLECTIONB =
        "<collection compressed='true' name='CollectionB'>" +
            "<filer class='org.apache.xindice.core.filer.BTreeFiler'/>" +
            "<indexes/>" +
        "</collection>";

    public static final String COLLECTIONC =
        "<collection compressed='true' name='CollectionC' inline-metadata='true'>" +
            "<filer class='org.apache.xindice.core.filer.BTreeFiler'/>" +
            "<indexes/>" +
        "</collection>";

	private Database db;
    private Configuration configurationA;
    private Configuration configurationB;
    private Configuration configurationC;

    public DatabaseTest(String name) {
        super(name);
    }

    public void setUp() throws Exception {
        db = new Database();
        db.setConfig(new Configuration(DOMParser.toDocument(DATABASE)));
        configurationA = new Configuration(DOMParser.toDocument(COLLECTIONA));
        configurationB = new Configuration(DOMParser.toDocument(COLLECTIONB));
        configurationC = new Configuration(DOMParser.toDocument(COLLECTIONC));
    }

    public void testCreateCollectionNullName() throws Exception {
        try {
            db.createCollection(null, configurationA);
            fail("DBException expected");
        } catch (DBException e) {
            // Ok
        }
    }

    public void testCreateCollectionEmptyName() throws Exception {
        try {
            db.createCollection("", configurationA);
            fail("DBException expected");
        } catch (DBException e) {
            // Ok
        }
    }

    public void testCreateCollectionNullConfig() throws Exception {
        try {
            db.createCollection("name", null);
            fail("DBException expected");
        } catch (DBException e) {
            // Ok
        }
    }

    public void testCreateCollection() throws Exception {
        Collection collection = db.createCollection("CollectionA", configurationA);
        assertNotNull("Collection expected", collection);
        assertEquals("Collection name", "CollectionA", collection.getName());
    }

    public void testCreateDuplicateCollection() throws Exception {
        db.createCollection("CollectionA", configurationA);
        try {
            db.createCollection("CollectionA", configurationA);
            fail("DBException expected");
        } catch (DBException e) {
            // Ok
        }
    }

    public void testCreateCollectionWrongName() throws Exception {
        try {
            // Name does not match name in configuration
            db.createCollection("somepath", configurationA);
            fail("Exception expected");
        } catch (DBException e) {
            // Expected
        }
    }

    /**
     * Multithreaded getCollection test
     */
    public void testMTGetCollection() throws Exception {
        Collection a = db.createCollection("CollectionA", configurationA);
        Collection b = db.createCollection("CollectionB", configurationB);
        Collection c = db.createCollection("CollectionC", configurationC);

        final int THREADS = 20;
        final int ITERATIONS = 50;

        final boolean[] result = new boolean[]{ true };
        Thread[] threads = new Thread[THREADS];
        for (int i = 0; i < THREADS; i++) {
            threads[i] = new Thread() {
                public void run() {
                    for (int ii = 0; ii < ITERATIONS && result[0]; ii++) {
                        Collection c;
                        try {
                            c = db.getCollection("CollectionA");
                            if (!c.getName().equals("CollectionA")) {
                                System.out.println("Expected CollectionA, got " + c.getName());
                                result[0] = false;
                            }
                            c = db.getCollection("CollectionB");
                            if (!c.getName().equals("CollectionB")) {
                                System.out.println("Expected CollectionB, got " + c.getName());
                                result[0] = false;
                            }
                            c = db.getCollection("CollectionC");
                            if (!c.getName().equals("CollectionC")) {
                                System.out.println("Expected CollectionC, got " + c.getName());
                                result[0] = false;
                            }
                        } catch (Exception e) {
                            e.printStackTrace();
                            result[0] = false;
                            fail();
                        }
                    }
                }
            };
            threads[i].setName("DatabaseTest" + i);
        }

        // Start all the threads at once
        for (int i = 0; i < THREADS; i++) {
            threads[i].start();
        }
        Thread.sleep(1000);
        for (int i = 0; i < THREADS; i++) {
            threads[i].join();
        }

        db.dropCollection(a);
        db.dropCollection(b);
        db.dropCollection(c);

        assertTrue("Test Failed", result[0]);
    }

    public void tearDown() throws Exception {
        try {
            db.dropCollection(db.getCollection("CollectionA"));
        } catch (Exception e) {}
        try {
            db.dropCollection(db.getCollection("CollectionB"));
        } catch (Exception e) {}
        try {
            db.dropCollection(db.getCollection("CollectionC"));
        } catch (Exception e) {}
        db.close();
    }
}
