/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.xindice.webadmin;

import org.apache.xindice.webadmin.webdav.components.Move;
import org.apache.xindice.webadmin.webdav.WebdavStatus;
import org.apache.xindice.xml.dom.DOMParser;
import org.apache.xindice.core.Collection;
import org.apache.xindice.util.Configuration;
import org.w3c.dom.Document;

public class MoveTest extends MethodSetup {

    public void setUp() throws Exception {
        super.setUp();
        method = new Move();
    }

    public void testMoveNullCollection() throws Exception {
        // source collection was not found
        method.execute(request, response, new Location("/unknown"));
        assertEquals(WebdavStatus.SC_NOT_FOUND, httpResponse.getStatus());
    }

    public void testMoveNoDestination() throws Exception {
        // request does not have a Destination header
        method.execute(request, response, new Location(collection.getCanonicalName()));
        assertEquals(WebdavStatus.SC_BAD_REQUEST, httpResponse.getStatus());

        method.execute(request, response, new Location(collection.getCanonicalDocumentName("test")));
        assertEquals(WebdavStatus.SC_BAD_REQUEST, httpResponse.getStatus());
    }

    public void testMoveSamePathResource() throws Exception {
        httpRequest.setPathInfo("/db/test");
        httpRequest.addHeader("Destination", "/db/test");
        method.execute(request, response, new Location(collection.getCanonicalDocumentName("test")));
        assertEquals(WebdavStatus.SC_FORBIDDEN, httpResponse.getStatus());
    }

    public void testMoveSamePathCollection() throws Exception {
        httpRequest.setPathInfo(collection.getCanonicalName());
        httpRequest.addHeader("Destination", collection.getCanonicalName() + "/new");
        method.execute(request, response, new Location(collection.getCanonicalName()));
        assertEquals(WebdavStatus.SC_FORBIDDEN, httpResponse.getStatus());

        httpRequest.setPathInfo(collection.getCanonicalName());
        httpRequest.addHeader("Destination", collection.getCanonicalName());
        method.execute(request, response, new Location(collection.getCanonicalName()));
        assertEquals(WebdavStatus.SC_FORBIDDEN, httpResponse.getStatus());
    }

    public void testMoveResourceNullDestCollection() throws Exception {
        // destination collection does not exist
        httpRequest.addHeader("Destination", collection.getCanonicalName() + "/test/resource");

        String xml = "<test document='true'>This is the test XML</test>";
        Document document = DOMParser.toDocument(xml);

        collection.setDocument("resource", document);

        method.execute(request, response, new Location(collection.getCanonicalDocumentName("resource")));
        assertEquals(WebdavStatus.SC_CONFLICT, httpResponse.getStatus());
    }

    public void testMoveResource() throws Exception {
        String name = "new";
        Collection child = collection.createCollection(name, new Configuration(
                DOMParser.toDocument(
                        "<collection compressed='true' name='" + name + "' inline-metadata='true'>" +
                        "<filer class='org.apache.xindice.core.filer.BTreeFiler' />" +
                        "</collection>"), false
        ));

        httpRequest.addHeader("Destination", collection.getCanonicalName() + "/" + name + "/resource");

        String xml = "<test document='true'>This is the test XML</test>";
        Document document = DOMParser.toDocument(xml);

        String docName = "resource";
        collection.setDocument(docName, document);

        method.execute(request, response, new Location(collection.getCanonicalDocumentName(docName)));
        assertEquals(WebdavStatus.SC_CREATED, httpResponse.getStatus());
        assertNotNull(child.getDocument(docName));
        assertEquals(document.toString(), child.getDocument(docName).toString());
        assertNull(collection.getDocument(docName));
    }

    public void testMoveResourceNoOverwrite() throws Exception {
        String name = "new";
        Collection child = collection.createCollection(name, new Configuration(
                DOMParser.toDocument(
                        "<collection compressed='true' name='" + name + "' inline-metadata='true'>" +
                        "<filer class='org.apache.xindice.core.filer.BTreeFiler' />" +
                        "</collection>"), false
        ));


        httpRequest.addHeader("Destination", collection.getCanonicalName() + "/" + name + "/resource");
        httpRequest.addHeader("Overwrite", "F");

        String xml1 = "<test1 document='true'>This is the test XML</test1>";
        Document document1 = DOMParser.toDocument(xml1);

        String xml2 = "<test2 document='true'>This is the test XML</test2>";
        Document document2 = DOMParser.toDocument(xml2);

        String docName = "resource";
        collection.setDocument(docName, document1);
        child.setDocument(docName, document2);

        method.execute(request, response, new Location(collection.getCanonicalDocumentName(docName)));
        assertEquals(WebdavStatus.SC_PRECONDITION_FAILED, httpResponse.getStatus());
        assertNotNull(collection.getDocument(docName));
        assertEquals(document2.toString(), child.getDocument(docName).toString());
    }

    public void testMoveCollection() throws Exception {
        String childName = "col1";
        Collection col1 = collection.createCollection(childName, new Configuration(
                DOMParser.toDocument(
                        "<collection compressed='true' name='" + childName + "' inline-metadata='true'>" +
                        "<filer class='org.apache.xindice.core.filer.BTreeFiler' />" +
                        "</collection>"), false
        ));

        int docs = 50;
        for (int i = 0; i < docs; i++) {
            String xml = "<test document='true'>This is the test XML " + i + "</test>";
            Document document = DOMParser.toDocument(xml);

            String docName = "resource" + i;
            collection.setDocument(docName, document);
        }

        for (int i = 0; i < docs; i++) {
            String xml = "<test>This is the inner collection test XML " + i + "</test>";
            Document document = DOMParser.toDocument(xml);

            String docName = "resource" + i;
            col1.setDocument(docName, document);
        }

        Collection destParent = null;
        try {
            String destName = "new";
            destParent = db.createCollection(destName, new Configuration(
                    DOMParser.toDocument(
                            "<collection compressed='true' name='" + destName + "' inline-metadata='true'>" +
                            "<filer class='org.apache.xindice.core.filer.BTreeFiler' />" +
                            "</collection>"), false
            ));

            String destColName = "dest";
            httpRequest.addHeader("Destination", destParent.getCanonicalName() + "/" + destColName);

            method.execute(request, response, new Location(collection.getCanonicalName()));
            assertEquals(WebdavStatus.SC_CREATED, httpResponse.getStatus());
            Collection destCollection = destParent.getCollection(destColName);
            assertNotNull(destCollection);
            assertEquals(destCollection.getDocumentCount(), 50);

            Collection child = destCollection.getCollection(childName);
            assertNotNull(child);
            assertEquals(child.getDocumentCount(), 50);

            assertNull(db.getCollection(collection.getName()));
        } finally {
            if (destParent != null) {
                db.dropCollection(destParent);
            }
        }
    }
}
